"""Unit tests for reviewboard.admin.forms.change_form.ChangeFormField."""

from django import forms
from django.contrib.admin.helpers import AdminField, AdminReadonlyField
from django.contrib.admin.widgets import FilteredSelectMultiple
from django.contrib.auth.models import User
from django.utils.safestring import SafeText

from reviewboard.admin import admin_site
from reviewboard.admin.form_widgets import (RelatedGroupWidget,
                                            RelatedRepositoryWidget,
                                            RelatedUserWidget)
from reviewboard.admin.forms.change_form import (ChangeFormField,
                                                 ChangeFormRow)
from reviewboard.reviews.models import Group, ReviewRequest
from reviewboard.scmtools.models import Repository
from reviewboard.testing.testcase import TestCase


class ChangeFormFieldTests(TestCase):
    """Unit tests for ChangeFormField."""

    def setUp(self):
        super(ChangeFormFieldTests, self).setUp()

        request = self.create_http_request()
        user = self.create_user()

        self.model_admin = admin_site.get_model_admin(User)
        self.form = self.model_admin.get_form(request, user)(instance=user)

    def test_init(self):
        """Testing ChangeFormField.__init__"""
        row = ChangeFormRow(self.form, field='username')
        field = ChangeFormField(form_row=row,
                                admin_field=AdminField(self.form,
                                                       field='username',
                                                       is_first=True))

        self.assertEqual(field.classes, 'rb-c-form-field -is-required')
        self.assertTrue(field.is_first)
        self.assertFalse(field.has_field_first)
        self.assertFalse(field.is_checkbox)
        self.assertFalse(field.is_readonly)
        self.assertFalse(field.show_errors)

    def test_init_with_read_only(self):
        """Testing ChangeFormField.__init__ with read-only field"""
        row = ChangeFormRow(self.form, field='username')
        field = ChangeFormField(
            form_row=row,
            admin_field=AdminReadonlyField(self.form,
                                           field='username',
                                           is_first=True,
                                           model_admin=self.model_admin))

        self.assertEqual(field.classes, 'rb-c-form-field -is-read-only')
        self.assertTrue(field.is_first)
        self.assertTrue(field.is_readonly)
        self.assertFalse(field.has_field_first)
        self.assertFalse(field.is_checkbox)
        self.assertFalse(field.show_errors)

    def test_init_with_errors(self):
        """Testing ChangeFormField.__init__ with errors"""
        self.form.errors['username'] = self.form.error_class(['Stuff bad'])
        row = ChangeFormRow(self.form, field='username')
        field = ChangeFormField(
            form_row=row,
            admin_field=AdminField(self.form,
                                   field='username',
                                   is_first=True))

        self.assertEqual(field.classes,
                         'rb-c-form-field -is-required -has-errors')
        self.assertTrue(field.show_errors)
        self.assertHTMLEqual(field.errors,
                             '<ul class="errorlist"><li>Stuff bad</li></ul>')

    def test_init_with_checkbox(self):
        """Testing ChangeFormField.__init__ with checkbox"""
        row = ChangeFormRow(self.form, field='is_active')
        field = ChangeFormField(
            form_row=row,
            admin_field=AdminField(self.form,
                                   field='is_active',
                                   is_first=True))

        self.assertEqual(field.classes, 'rb-c-form-field -has-input-first')
        self.assertTrue(field.has_field_first)
        self.assertTrue(field.is_checkbox)

    def test_init_with_multi_choice(self):
        """Testing ChangeFormField.__init__ with ModelMultipleChoiceField"""
        class MyForm(forms.Form):
            field = forms.ModelMultipleChoiceField(
                queryset=ReviewRequest.objects.all(),
                required=False)

        form = MyForm()
        row = ChangeFormRow(form, field='field')
        field = ChangeFormField(
            form_row=row,
            admin_field=AdminField(form,
                                   field='field',
                                   is_first=True))

        self.assertEqual(field.classes, 'rb-c-form-field')
        self.assertIsInstance(form.fields['field'].widget,
                              FilteredSelectMultiple)

    def test_init_with_multi_choice_user(self):
        """Testing ChangeFormField.__init__ with ModelMultipleChoiceField with
        User model
        """
        class MyForm(forms.Form):
            field = forms.ModelMultipleChoiceField(
                queryset=User.objects.all(),
                required=False)

        form = MyForm()
        row = ChangeFormRow(form, field='field')
        field = ChangeFormField(
            form_row=row,
            admin_field=AdminField(form,
                                   field='field',
                                   is_first=True))

        self.assertEqual(field.classes, 'rb-c-form-field')
        self.assertIsInstance(form.fields['field'].widget, RelatedUserWidget)

    def test_init_with_multi_choice_repository(self):
        """Testing ChangeFormField.__init__ with ModelMultipleChoiceField with
        Repository model
        """
        class MyForm(forms.Form):
            field = forms.ModelMultipleChoiceField(
                queryset=Repository.objects.all(),
                required=False)

        form = MyForm()
        row = ChangeFormRow(form, field='field')
        field = ChangeFormField(
            form_row=row,
            admin_field=AdminField(form,
                                   field='field',
                                   is_first=True))

        self.assertEqual(field.classes, 'rb-c-form-field')
        self.assertIsInstance(form.fields['field'].widget,
                              RelatedRepositoryWidget)

    def test_init_with_multi_choice_group(self):
        """Testing ChangeFormField.__init__ with ModelMultipleChoiceField with
        Group model
        """
        class MyForm(forms.Form):
            field = forms.ModelMultipleChoiceField(
                queryset=Group.objects.all(),
                required=False)

        form = MyForm()
        row = ChangeFormRow(form, field='field')
        field = ChangeFormField(
            form_row=row,
            admin_field=AdminField(form,
                                   field='field',
                                   is_first=True))

        self.assertEqual(field.classes, 'rb-c-form-field')
        self.assertIsInstance(form.fields['field'].widget, RelatedGroupWidget)

    def test_label_tag(self):
        """Testing ChangeFormField.label_tag"""
        row = ChangeFormRow(self.form, field='username')
        field = ChangeFormField(form_row=row,
                                admin_field=AdminField(self.form,
                                                       field='username',
                                                       is_first=True))
        html = field.label_tag()

        self.assertIsInstance(html, SafeText)
        self.assertHTMLEqual(
            html,
            '<label class="rb-c-form-field__label"'
            ' for="id_username">Username:</label>')

    def test_label_tag_with_read_only_field(self):
        """Testing ChangeFormField.label_tag with read-only field"""
        row = ChangeFormRow(self.form, field='username')
        field = ChangeFormField(
            form_row=row,
            admin_field=AdminReadonlyField(self.form,
                                           field='username',
                                           is_first=True,
                                           model_admin=self.model_admin))
        html = field.label_tag()

        self.assertIsInstance(html, SafeText)
        self.assertHTMLEqual(
            html,
            '<label class="rb-c-form-field__label">Username:</label>')

    def test_label_tag_with_checkbox_field(self):
        """Testing ChangeFormField.label_tag with checkbox field"""
        row = ChangeFormRow(self.form, field='is_active')
        field = ChangeFormField(
            form_row=row,
            admin_field=AdminField(self.form,
                                   field='is_active',
                                   is_first=True))
        html = field.label_tag()

        self.assertIsInstance(html, SafeText)
        self.assertHTMLEqual(
            html,
            '<label class="rb-c-form-field__label" for="id_is_active">'
            'Active</label>')

    def test_render(self):
        """Testing ChangeFormField.render"""
        row = ChangeFormRow(self.form, field='is_active')
        field = ChangeFormField(form_row=row,
                                admin_field=AdminField(self.form,
                                                       field='is_active',
                                                       is_first=True))
        html = field.render()

        self.assertIsInstance(html, SafeText)
        self.assertHTMLEqual(
            html,
            '<input id="id_is_active" name="is_active" type="checkbox"'
            ' checked>')

    def test_render_with_read_only_field(self):
        """Testing ChangeFormField.render with read-only field"""
        row = ChangeFormRow(self.form, field='username')
        field = ChangeFormField(
            form_row=row,
            admin_field=AdminReadonlyField(self.form,
                                           field='username',
                                           is_first=True,
                                           model_admin=self.model_admin))
        html = field.render()

        self.assertIsInstance(html, SafeText)
        self.assertHTMLEqual(
            html,
            '<div class="rb-c-form-field__readonly-value">test-user</div>')



"""Hosting service support for GitHub."""

from __future__ import annotations

import hashlib
import hmac
import json
import logging
import re
from collections import defaultdict
from typing import Optional, TYPE_CHECKING
from urllib.parse import urljoin

from django import forms
from django.core.cache import cache
from django.core.exceptions import ObjectDoesNotExist
from django.http import HttpResponse, HttpResponseBadRequest
from django.template.loader import render_to_string
from django.urls import path
from django.utils.text import format_lazy
from django.utils.translation import gettext, gettext_lazy as _
from django.views.decorators.http import require_POST
from housekeeping import deprecate_non_keyword_only_args

from reviewboard import get_manual_url
from reviewboard.admin.server import build_server_url, get_server_url
from reviewboard.deprecation import RemovedInReviewBoard10_0Warning
from reviewboard.hostingsvcs.base.bug_tracker import BaseBugTracker
from reviewboard.hostingsvcs.base.client import HostingServiceClient
from reviewboard.hostingsvcs.base.hosting_service import BaseHostingService
from reviewboard.hostingsvcs.errors import (AuthorizationError,
                                            HostingServiceError,
                                            InvalidPlanError,
                                            RepositoryError)
from reviewboard.hostingsvcs.base.forms import (
    BaseHostingServiceAuthForm,
    BaseHostingServiceRepositoryForm,
)
from reviewboard.hostingsvcs.hook_utils import (close_all_review_requests,
                                                get_git_branch_name,
                                                get_repository_for_hook,
                                                get_review_request_id)
from reviewboard.hostingsvcs.repository import RemoteRepository
from reviewboard.hostingsvcs.utils.paginator import (APIPaginator,
                                                     ProxyPaginator)
from reviewboard.scmtools.core import Branch, Commit
from reviewboard.scmtools.crypto_utils import (decrypt_password,
                                               encrypt_password)
from reviewboard.scmtools.errors import FileNotFoundError, SCMError
from reviewboard.site.urlresolvers import local_site_reverse

if TYPE_CHECKING:
    from urllib.error import URLError

    from django.http import HttpRequest

    from reviewboard.hostingsvcs.base.bug_tracker import BugInfo
    from reviewboard.hostingsvcs.base.http import HostingServiceHTTPRequest
    from reviewboard.scmtools.models import Repository


logger = logging.getLogger(__name__)


#: A list of the scopes that Review Board requires for classic tokens.
#:
#: Fine-grained Personal Access Tokens use a different permission model and
#: don't return scopes via the ``X-OAuth-Scopes`` header, so this list only
#: applies to classic Personal Access Tokens.
_REQUIRED_SCOPES = ['repo']


def _is_fine_grained_pat(
    token: str,
) -> bool:
    """Return whether a GitHub token is a fine-grained Personal Access Token.

    Fine-grained Personal Access Tokens are prefixed with ``github_pat_``.
    Everything else is treated as classic and validated via the X-Oauth-Scopes
    header.

    Version Added:
        8.0

    Args:
        token (str):
            The token to check.

    Returns:
        bool:
        ``True`` if the token is a fine-grained PAT, ``False`` otherwise.
    """
    return token.startswith('github_pat_')


class GitHubAuthForm(BaseHostingServiceAuthForm):
    """Form for authenticating to GitHub."""

    class Meta:
        """Metadata for the GitHubAuthForm."""

        labels = {
            'hosting_account_username': _('GitHub Username'),
            'hosting_account_password': _('Personal Access Token'),
        }

        help_texts = {
            'hosting_account_username': _(
                'Your GitHub username. This must <em>not</em> be your '
                'e-mail address!'
            ),
            'hosting_account_password': format_lazy(
                _(
                    'This must be a fine-grained access token (recommended) '
                    'or a classic access token. See the '
                    '<a href="{docs_url}" target="_blank">documentation</a> '
                    'for details on the trade-offs and how to choose the '
                    'right token for your needs.'
                ),
                docs_url=(
                    get_manual_url() +
                    'admin/configuration/repositories/github/'
                ),
            ),
        }


class GitHubPublicForm(BaseHostingServiceRepositoryForm):
    """Sub-form for public repositories owned by a user."""

    github_public_repo_name = forms.CharField(
        label=_('Repository name'),
        max_length=64,
        required=True,
        widget=forms.TextInput(attrs={'size': '60'}),
        help_text=_('The name of the repository. This is the '
                    '<code>&lt;repo_name&gt;</code> in '
                    '<code>http://github.com/&lt;username&gt;/'
                    '&lt;repo_name&gt;/</code>'))


class GitHubPrivateForm(BaseHostingServiceRepositoryForm):
    """Sub-form for private repositories owned by a user."""

    github_private_repo_name = forms.CharField(
        label=_('Repository name'),
        max_length=64,
        required=True,
        widget=forms.TextInput(attrs={'size': '60'}),
        help_text=_('The name of the repository. This is the '
                    '<code>&lt;repo_name&gt;</code> in '
                    '<code>http://github.com/&lt;username&gt;/'
                    '&lt;repo_name&gt;/</code>'))


class GitHubPublicOrgForm(BaseHostingServiceRepositoryForm):
    """Sub-form for public repositories owned by an organization."""

    github_public_org_name = forms.CharField(
        label=_('Organization name'),
        max_length=64,
        required=True,
        widget=forms.TextInput(attrs={'size': '60'}),
        help_text=_('The name of the organization. This is the '
                    '<code>&lt;org_name&gt;</code> in '
                    '<code>http://github.com/&lt;org_name&gt;/'
                    '&lt;repo_name&gt;/</code>'))

    github_public_org_repo_name = forms.CharField(
        label=_('Repository name'),
        max_length=64,
        required=True,
        widget=forms.TextInput(attrs={'size': '60'}),
        help_text=_('The name of the repository. This is the '
                    '<code>&lt;repo_name&gt;</code> in '
                    '<code>http://github.com/&lt;org_name&gt;/'
                    '&lt;repo_name&gt;/</code>'))


class GitHubPrivateOrgForm(BaseHostingServiceRepositoryForm):
    """Sub-form for private repositories owned by an organization."""

    github_private_org_name = forms.CharField(
        label=_('Organization name'),
        max_length=64,
        required=True,
        widget=forms.TextInput(attrs={'size': '60'}),
        help_text=_('The name of the organization. This is the '
                    '<code>&lt;org_name&gt;</code> in '
                    '<code>http://github.com/&lt;org_name&gt;/'
                    '&lt;repo_name&gt;/</code>'))

    github_private_org_repo_name = forms.CharField(
        label=_('Repository name'),
        max_length=64,
        required=True,
        widget=forms.TextInput(attrs={'size': '60'}),
        help_text=_('The name of the repository. This is the '
                    '<code>&lt;repo_name&gt;</code> in '
                    '<code>http://github.com/&lt;org_name&gt;/'
                    '&lt;repo_name&gt;/</code>'))


class GitHubAPIPaginator(APIPaginator):
    """Paginates over GitHub API list resources.

    This is returned by some GitHubClient functions in order to handle
    iteration over pages of results, without resorting to fetching all
    pages at once or baking pagination into the functions themselves.
    """
    start_query_param = 'page'
    per_page_query_param = 'per_page'

    LINK_RE = re.compile(r'\<(?P<url>[^>]+)\>; rel="(?P<rel>[^"]+)",? *')

    def fetch_url(self, url):
        """Fetches the page data from a URL."""
        rsp = self.client.http_get(url)

        # Find all the links in the Link header and key off by the link
        # name ('prev', 'next', etc.).
        link_header = rsp.get_header('Link', '')

        links = {
            m.group('rel'): m.group('url')
            for m in self.LINK_RE.finditer(link_header)
        }

        return {
            'data': rsp.json,
            'response': rsp,
            'headers': rsp.headers,
            'prev_url': links.get('prev'),
            'next_url': links.get('next'),
        }


class GitHubClient(HostingServiceClient):
    RAW_MIMETYPE = 'application/vnd.github.v3.raw'

    def __init__(self, hosting_service):
        super(GitHubClient, self).__init__(hosting_service)
        self.account = hosting_service.account

    def get_http_credentials(self, account, username=None, password=None,
                             **kwargs):
        """Return credentials used to authenticate with GitHub.

        Unless an explicit username and password is provided, this will
        use the ones stored for the account.

        Args:
            account (reviewboard.hostingsvcs.models.HostingServiceAccount):
                The stored authentication data for the service.

            username (unicode, optional):
                An explicit username passed by the caller. This will override
                the data stored in the account, if both a username and
                password are provided.

            password (unicode, optional):
                An explicit password passed by the caller. This will override
                the data stored in the account, if both a username and
                password are provided.

            **kwargs (dict, unused):
                Additional keyword arguments passed in when making the HTTP
                request.

        Returns:
            dict:
            A dictionary of credentials for the request.
        """
        if username is None and password is None:
            if 'personal_token' in account.data:
                username = account.username
                password = decrypt_password(account.data['personal_token'])
            elif ('authorization' in account.data and
                  'token' in account.data['authorization']):
                username = account.username
                password = account.data['authorization']['token']

        if username is not None and password is not None:
            return {
                'username': username,
                'password': password,
            }

        return {}

    def process_http_response(self, response):
        """Process an HTTP response and return a result.

        Args:
            response (reviewboard.hostingsvcs.base.http.
                      HostingServiceHTTPResponse):
                The response to process.

        Returns:
            reviewboard.hostingsvcs.base.http.HostingServiceHTTPResponse:
            The resulting response.
        """
        rate_limit_remaining = response.get_header('X-RateLimit-Remaining')

        try:
            if (rate_limit_remaining is not None and
                int(rate_limit_remaining) <= 100):
                logger.warning('GitHub rate limit for %s is down to %s',
                               self.account.username, rate_limit_remaining)
        except ValueError:
            pass

        return response

    def process_http_error(
        self,
        request: HostingServiceHTTPRequest,
        e: URLError,
    ) -> None:
        """Process an HTTP error, possibly raising a result.

        This will look at the error, possibly raising a more suitable exception
        in its place. It checks for SSL verification failures, bad credentials,
        and GitHub error payloads.

        Args:
            request (reviewboard.hostingsvcs.base.http.
                     HostingServiceHTTPRequest):
                The request that resulted in an error.

            e (urllib2.URLError):
                The error to process.

        Raises:
            reviewboard.hostingsvcs.errors.AuthorizationError:
                The repository credentials are invalid.

            reviewboard.hostingsvcs.errors.HostingServiceError:
                There was an error with the request. Details are in the
                response.

            reviewboard.scmtools.errors.UnverifiedCertificateError:
                The SSL certificate was not able to be verified.
        """
        super().process_http_error(request, e)

        try:
            data = e.read()  # type: ignore
            rsp = json.loads(data.decode('utf-8'))
        except Exception:
            rsp = None

        http_code: Optional[int] = getattr(e, 'code', None)

        if rsp and 'message' in rsp:
            message = rsp['message']

            if http_code == 401:
                raise AuthorizationError(message, http_code=http_code)

            raise HostingServiceError(message, http_code=http_code)
        else:
            raise HostingServiceError(str(e), http_code=http_code)

    #
    # Higher-level API methods
    #

    def api_get_list(self, url, start=None, per_page=None, *args, **kwargs):
        """Perform an HTTP GET to a GitHub API and returns a paginator.

        This returns a GitHubAPIPaginator that's used to iterate over the
        pages of results. Each page contains information on the data and
        headers from that given page.

        The ``start`` and ``per_page`` parameters can be used to control
        where pagination begins and how many results are returned per page.
        ``start`` is a 0-based index representing a page number.
        """
        if start is not None:
            # GitHub uses 1-based indexing, so add one.
            start += 1

        return GitHubAPIPaginator(
            client=self,
            url=url,
            start=start,
            per_page=per_page)

    def api_get_blob(self, repo_api_url, path, sha):
        """Return the contents of a file using the GitHub API.

        Args:
            repo_api_url (unicode):
                The absolute URL for the base repository API.

            path (unicode):
                The path of the file within the repository.

            sha (unicode):
                The SHA1 of the file within the repository.

        Returns:
            bytes:
            The contents of the file.

        Raises:
            reviewboard.scmtools.errors.FileNotFoundError:
                The file could not be found or the API could not be accessed.
        """
        try:
            return self.http_get(
                url='%s/git/blobs/%s' % (repo_api_url, sha),
                headers={
                    'Accept': self.RAW_MIMETYPE,
                }).data
        except HostingServiceError:
            raise FileNotFoundError(path, sha)

    def api_get_commits(self, repo_api_url, branch=None, start=None):
        url = '%s/commits' % repo_api_url

        # Note that we don't always use the branch, since the GitHub API
        # doesn't support limiting by branch *and* starting at a SHA. So, the
        # branch argument can be safely ignored if a sha is provided.
        start = start or branch

        if start:
            url += '?sha=%s' % start

        try:
            return self.http_get(url).json
        except Exception as e:
            logger.warning('Failed to fetch commits from %s: %s',
                           url, e, exc_info=True)
            raise SCMError(str(e))

    def api_get_compare_commits(self, repo_api_url, parent_revision, revision):
        # If the commit has a parent commit, use GitHub's "compare two commits"
        # API to get the diff. Otherwise, fetch the commit itself.
        if parent_revision:
            url = '%s/compare/%s...%s' % (repo_api_url, parent_revision,
                                          revision)
        else:
            url = '%s/commits/%s' % (repo_api_url, revision)

        try:
            comparison = self.http_get(url).json
        except Exception as e:
            logger.warning('Failed to fetch commit comparison from %s: %s',
                           url, e, exc_info=True)
            raise SCMError(str(e))

        if parent_revision:
            tree_sha = comparison['base_commit']['commit']['tree']['sha']
        else:
            tree_sha = comparison['commit']['tree']['sha']

        return comparison['files'], tree_sha

    def api_get_heads(self, repo_api_url):
        url = '%s/git/refs/heads' % repo_api_url

        try:
            rsp = self.http_get(url).json
            return [ref for ref in rsp if ref['ref'].startswith('refs/heads/')]
        except Exception as e:
            logger.warning('Failed to fetch commits from %s: %s',
                           url, e, exc_info=True)
            raise SCMError(str(e))

    def api_get_issue(self, repo_api_url, issue_id):
        url = '%s/issues/%s' % (repo_api_url, issue_id)

        try:
            return self.http_get(url).json
        except Exception as e:
            logger.warning('GitHub: Failed to fetch issue from %s: %s',
                           url, e, exc_info=True)
            raise SCMError(str(e))

    def api_get_remote_repositories(self, api_url, owner, owner_type,
                                    filter_type=None, start=None,
                                    per_page=None):
        url = api_url

        if owner_type == 'organization':
            url += 'orgs/%s/repos' % owner
        elif owner_type == 'user':
            if owner == self.account.username:
                # All repositories belonging to an authenticated user.
                url += 'user/repos'
            else:
                # Only public repositories for the user.
                url += 'users/%s/repos' % owner
        else:
            raise ValueError(
                "owner_type must be 'organization' or 'user', not %r'"
                % owner_type)

        if filter_type:
            url += '?type=%s' % (filter_type or 'all')

        return self.api_get_list(url,
                                 start=start,
                                 per_page=per_page)

    def api_get_remote_repository(self, api_url, owner, repository_id):
        try:
            return self.http_get(
                '%srepos/%s/%s' % (api_url, owner, repository_id)).json
        except HostingServiceError as e:
            if e.http_code == 404:
                return None
            else:
                raise

    def api_get_tree(self, repo_api_url, sha, recursive=False):
        url = '%s/git/trees/%s' % (repo_api_url, sha)

        if recursive:
            url += '?recursive=1'

        try:
            return self.http_get(url).json
        except Exception as e:
            logger.warning('Failed to fetch tree from %s: %s',
                           url, e, exc_info=True)
            raise SCMError(str(e))


class GitHubHookViews(object):
    """Container class for hook views."""

    @staticmethod
    @require_POST
    def post_receive_hook_close_submitted(
        request: HttpRequest,
        local_site_name: (str | None) = None,
        repository_id: (int | None) = None,
        hosting_service_id: (str | None) = None,
    ) -> HttpResponse:
        """Close review requests as submitted automatically after a push.

        Args:
            request (django.http.HttpRequest):
                The request from the Bitbucket webhook.

            local_site_name (str, optional):
                The local site name, if available.

            repository_id (int, optional):
                The pk of the repository, if available.

            hosting_service_id (str, optional):
                The name of the hosting service.

        Returns:
            django.http.HttpResponse:
            A response for the request.
        """
        hook_event = request.META.get('HTTP_X_GITHUB_EVENT')

        if hook_event == 'ping':
            # GitHub is checking that this hook is valid, so accept the request
            # and return.
            return HttpResponse()
        elif hook_event != 'push':
            return HttpResponseBadRequest(
                'Only "ping" and "push" events are supported.')

        assert repository_id is not None
        assert hosting_service_id is not None

        repository = get_repository_for_hook(
            repository_id=repository_id,
            hosting_service_id=hosting_service_id,
            local_site_name=local_site_name)

        # Validate the hook against the stored UUID.
        m = hmac.new(repository.get_or_create_hooks_uuid().encode('utf-8'),
                     request.body, hashlib.sha1)

        sig_parts = request.META.get('HTTP_X_HUB_SIGNATURE').split('=')

        if sig_parts[0] != 'sha1' or len(sig_parts) != 2:
            # We don't know what this is.
            return HttpResponseBadRequest('Unsupported HTTP_X_HUB_SIGNATURE')

        if m.hexdigest() != sig_parts[1]:
            return HttpResponseBadRequest('Bad signature.')

        try:
            payload = json.loads(request.body.decode('utf-8'))
        except ValueError as e:
            logger.error('The payload is not in JSON format: %s', e)
            return HttpResponseBadRequest('Invalid payload format')

        server_url = get_server_url(request=request)
        review_request_id_to_commits = \
            GitHubHookViews._get_review_request_id_to_commits_map(
                payload, server_url, repository)

        if review_request_id_to_commits:
            close_all_review_requests(
                review_request_id_to_commits=review_request_id_to_commits,
                local_site_name=local_site_name,
                repository=repository,
                hosting_service_id=hosting_service_id)

        return HttpResponse()

    @staticmethod
    def _get_review_request_id_to_commits_map(payload, server_url, repository):
        """Return a mapping of review request ID to a list of commits.

        If a commit's commit message does not contain a review request ID,
        we append the commit to the key None.

        Args:
            payload (dict):
                The decoded webhook payload.

            server_url (unicode):
                The URL of the Review Board server.

            repository (reviewboard.scmtools.models.Repository):
                The repository object.

        Returns:
            dict:
            A mapping from review request ID to a list of matching commits from
            the payload.
        """
        review_request_id_to_commits_map = defaultdict(list)

        ref_name = payload.get('ref')
        if not ref_name:
            return None

        branch_name = get_git_branch_name(ref_name)
        if not branch_name:
            return None

        commits = payload.get('commits', [])

        for commit in commits:
            commit_hash = commit.get('id')
            commit_message = commit.get('message')
            review_request_id = get_review_request_id(
                commit_message=commit_message,
                server_url=server_url,
                commit_id=commit_hash,
                repository=repository)

            review_request_id_to_commits_map[review_request_id].append(
                '%s (%s)' % (branch_name, commit_hash[:7]))

        return review_request_id_to_commits_map


class GitHub(BaseHostingService, BaseBugTracker):
    name = _('GitHub')
    hosting_service_id = 'github'
    plans = [
        ('public', {
            'name': _('Public'),
            'form': GitHubPublicForm,
            'repository_fields': {
                'Git': {
                    'path': 'git://github.com/%(hosting_account_username)s/'
                            '%(github_public_repo_name)s.git',
                    'mirror_path': 'git@github.com:'
                                   '%(hosting_account_username)s/'
                                   '%(github_public_repo_name)s.git',
                }
            },
            'bug_tracker_field': 'http://github.com/'
                                 '%(hosting_account_username)s/'
                                 '%(github_public_repo_name)s/'
                                 'issues#issue/%%s',
        }),
        ('public-org', {
            'name': _('Public Organization'),
            'form': GitHubPublicOrgForm,
            'repository_fields': {
                'Git': {
                    'path': 'git://github.com/%(github_public_org_name)s/'
                            '%(github_public_org_repo_name)s.git',
                    'mirror_path': 'git@github.com:%(github_public_org_name)s/'
                                   '%(github_public_org_repo_name)s.git',
                }
            },
            'bug_tracker_field': 'http://github.com/'
                                 '%(github_public_org_name)s/'
                                 '%(github_public_org_repo_name)s/'
                                 'issues#issue/%%s',
        }),
        ('private', {
            'name': _('Private'),
            'form': GitHubPrivateForm,
            'repository_fields': {
                'Git': {
                    'path': 'git@github.com:%(hosting_account_username)s/'
                            '%(github_private_repo_name)s.git',
                    'mirror_path': '',
                },
            },
            'bug_tracker_field': 'http://github.com/'
                                 '%(hosting_account_username)s/'
                                 '%(github_private_repo_name)s/'
                                 'issues#issue/%%s',
        }),
        ('private-org', {
            'name': _('Private Organization'),
            'form': GitHubPrivateOrgForm,
            'repository_fields': {
                'Git': {
                    'path': 'git@github.com:%(github_private_org_name)s/'
                            '%(github_private_org_repo_name)s.git',
                    'mirror_path': '',
                },
            },
            'bug_tracker_field': 'http://github.com/'
                                 '%(github_private_org_name)s/'
                                 '%(github_private_org_repo_name)s/'
                                 'issues#issue/%%s',
        }),
    ]

    auth_form = GitHubAuthForm

    needs_authorization = True
    supports_bug_trackers = True
    supports_post_commit = True
    supports_repositories = True
    supports_list_remote_repositories = True
    supported_scmtools = ['Git']

    has_repository_hook_instructions = True

    client_class = GitHubClient

    repository_url_patterns = [
        path('hooks/close-submitted/',
             GitHubHookViews.post_receive_hook_close_submitted,
             name='github-hooks-close-submitted')
    ]

    # This should be the prefix for every field on the plan forms.
    plan_field_prefix = 'github'

    #: A list of the scopes that Review Board requires.
    REQUIRED_SCOPES = _REQUIRED_SCOPES

    _ORG_ACCESS_SUPPORT_URL = (
        'https://beanbag.freshdesk.com/solution/articles/3000045767'
        '-granting-organization-access-on-github'
    )

    def get_api_url(self, hosting_url):
        """Returns the API URL for GitHub.

        This can be overridden to provide more advanced lookup (intended
        for the GitHub Enterprise support).
        """
        assert not hosting_url
        return 'https://api.github.com/'

    def get_plan_field(self, plan, plan_data, name):
        """Returns the value of a field for plan-specific data.

        This takes into account the plan type and hosting service ID.
        """
        key = '%s_%s_%s' % (self.plan_field_prefix, plan.replace('-', '_'),
                            name)
        return plan_data[key]

    @deprecate_non_keyword_only_args(RemovedInReviewBoard10_0Warning)
    def check_repository(
        self,
        *,
        plan: str,
        **kwargs,
    ) -> None:
        """Check the validity of a repository.

        This will perform an API request against GitHub to get
        information on the repository. This will throw an exception if
        the repository was not found, and return cleanly if it was found.

        Version Changed:
            8.0:
            Made arguments keyword-only.

        Args:
            plan (str):
                The ID of the plan that the repository is on.

            **kwargs (dict, unused):
                Additional keyword arguments passed by the repository form.

        Raises:
            reviewboard.hostingsvcs.errors.RepositoryError:
                The repository is not valid.
        """
        repo_api_url = self._get_repo_api_url_raw(
            self._get_repository_owner_raw(plan, kwargs),
            self._get_repository_name_raw(plan, kwargs))

        try:
            rsp = self.client.http_get(repo_api_url)
            repo_info = rsp.json
        except HostingServiceError as e:
            if e.http_code == 404:
                if plan in {'public', 'private'}:
                    raise RepositoryError(
                        gettext('A repository with this name was not found, '
                                'or your user may not own it.'))
                elif plan == 'public-org':
                    raise RepositoryError(
                        gettext('A repository with this organization or '
                                'name was not found.'))
                elif plan == 'private-org':
                    raise RepositoryError(
                        gettext('A repository with this organization or name '
                                'was not found, or your user may not have '
                                'access to it.'),
                        help_link=self._ORG_ACCESS_SUPPORT_URL,
                        help_link_text=gettext(
                            'Get help on granting access.'))

            raise

        if 'private' in repo_info:
            is_private = repo_info['private']

            if is_private and plan in {'public', 'public-org'}:
                raise RepositoryError(
                    gettext('This is a private repository, but you have '
                            'selected a public plan.'))
            elif not is_private and plan in {'private', 'private-org'}:
                raise RepositoryError(
                    gettext('This is a public repository, but you have '
                            'selected a private plan.'))

        # Make a request to an endpoint that requires the "Contents" permission
        # to verify that a fine-grained PAT has the correct permissions. We
        # use /branches rather than /git/refs/heads because the latter returns
        # HTTP 409 on empty repositories.
        try:
            self.client.http_get(f'{repo_api_url}/branches')
        except HostingServiceError as e:
            if (e.http_code == 403 and
                'Resource not accessible by personal access token' in str(e)):
                raise RepositoryError(gettext(
                    'Your token can access this repository\'s metadata but '
                    'cannot read its contents. For fine-grained Personal '
                    'Access Tokens, ensure "Contents: Read" is granted on '
                    'this repository.'
                ))

            raise

    @deprecate_non_keyword_only_args(RemovedInReviewBoard10_0Warning)
    def authorize(
        self,
        *,
        username: str | None,
        password: str | None,
        hosting_url: (str | None) = None,
        local_site_name: (str | None) = None,
        **kwargs,
    ) -> None:
        """Authorize an account for the hosting service.

        Version Changed:
            8.0:
            Made arguments keyword-only.

        Args:
            username (str):
                The username for the account.

            password (str):
                The Personal Access Token for the account.

            hosting_url (str, optional):
                The hosting URL for the service, if self-hosted.

            local_site_name (str, optional):
                The Local Site name, if any, that the account should be
                bound to.

            **kwargs (dict, unused):
                Extra keyword arguments containing values from the
                repository's configuration.

        Raises:
            reviewboard.hostingsvcs.errors.AuthorizationError:
                The credentials provided were not valid.
        """
        api_url = self.get_api_url(hosting_url)

        # Try to reach an API resource with the provided credentials.
        rsp = self.client.http_get(
            f'{api_url}user',
            username=username,
            password=password)

        # Fine-grained Personal Access Tokens don't return scopes via the
        # X-OAuth-Scopes header. The /user call above already validated the
        # token; per-repository permission errors will surface naturally
        # during check_repository or normal SCM operations.
        if not _is_fine_grained_pat(password or ''):
            # Check to make sure this classic token has all the necessary
            # scopes.
            scopes_header = rsp.get_header('x-oauth-scopes', '')
            assert scopes_header is not None

            token_scopes = set(scopes_header.split(', '))
            missing_scopes = set(self.REQUIRED_SCOPES) - token_scopes

            if missing_scopes:
                raise AuthorizationError(
                    _('This GitHub classic Personal Access Token must have '
                      'the following scopes enabled: %(scopes)s')
                    % {
                        'scopes': ', '.join(sorted(missing_scopes)),
                    })

        if 'authorization' in self.account.data:
            # This is an older GitHub linked account, which used the legacy
            # authorizations API to generate the token. This stopped being
            # supported in Review Board 3.0.18.
            del self.account.data['authorization']

        self.account.data['personal_token'] = encrypt_password(password)
        self.account.save()

    def is_authorized(self):
        """Return whether or not the account is currently authorized.

        This will check for both a configured Personal Access Token
        (introduced in Review Board 3.0.18) and a legacy
        authorizations-generated OAuth Token.

        Returns:
            bool:
            Whether or not the associated account is authorized.
        """
        account_data = self.account.data

        if account_data.get('personal_token'):
            # This is a newer linked account using a GitHub user's custom
            # Personal Access Token. Support for this was introduced in
            # Review Board 3.0.18.
            return True

        # Check for a legacy authorizations-generated API token.
        return ('authorization' in account_data and
                'token' in account_data['authorization'])

    def get_file(self, repository, path, revision, *args, **kwargs):
        repo_api_url = self._get_repo_api_url(repository)
        return self.client.api_get_blob(repo_api_url, path, revision)

    def get_file_exists(self, repository, path, revision, *args, **kwargs):
        try:
            repo_api_url = self._get_repo_api_url(repository)
            self.client.api_get_blob(repo_api_url, path, revision)
            return True
        except FileNotFoundError:
            return False

    def get_branches(self, repository):
        repo_api_url = self._get_repo_api_url(repository)
        refs = self.client.api_get_heads(repo_api_url)
        results = []

        # A lot of repositories are starting to use alternative names for
        # their mainline branch, and GitHub doesn't have a good way for us to
        # know which one is which. Until this is better defined, we'll still
        # prefer "master" when available, then look for "main", and finally
        # make sure that at least one branch is marked as default.
        master_ref = None
        main_ref = None

        for i, ref in enumerate(refs):
            name = ref['ref'][len('refs/heads/'):]
            results.append(Branch(id=name,
                                  commit=ref['object']['sha']))

            if name == 'master':
                master_ref = i
            elif name == 'main':
                main_ref = i

        if master_ref is not None:
            results[master_ref].default = True
        elif main_ref is not None:
            results[main_ref].default = True
        elif len(results) > 0:
            results[0].default = True

        return results

    def get_commits(self, repository, branch=None, start=None):
        repo_api_url = self._get_repo_api_url(repository)
        commits = self.client.api_get_commits(repo_api_url, branch=branch,
                                              start=start)

        results = []

        for item in commits:
            commit = Commit(
                author_name=item['commit']['author']['name'],
                id=item['sha'],
                date=item['commit']['committer']['date'],
                message=item['commit']['message'])

            if item['parents']:
                commit.parent = item['parents'][0]['sha']

            results.append(commit)

        return results

    def get_change(self, repository, revision):
        repo_api_url = self._get_repo_api_url(repository)

        # Step 1: fetch the commit itself that we want to review, to get
        # the parent SHA and the commit message. Hopefully this information
        # is still in cache so we don't have to fetch it again.
        commit = cache.get(repository.get_commit_cache_key(revision))

        if commit:
            author_name = commit.author_name
            date = commit.date
            parent_revision = commit.parent
            message = commit.message
        else:
            commit = self.client.api_get_commits(repo_api_url, revision)[0]

            author_name = commit['commit']['author']['name']
            date = commit['commit']['committer']['date']
            parent_revision = commit['parents'][0]['sha']
            message = commit['commit']['message']

        # Step 2: Get the diff and tree from the "compare commits" API
        files, tree_sha = self.client.api_get_compare_commits(
            repo_api_url, parent_revision, revision)

        # Step 3: fetch the tree for the original commit, so that we can get
        # full blob SHAs for each of the files in the diff.
        tree = self.client.api_get_tree(repo_api_url, tree_sha, recursive=True)

        file_shas = {
            f['path'].encode('utf-8'): f['sha'].encode('utf-8')
            for f in tree['tree']
        }

        diff = []

        for f in files:
            filename = f['filename'].encode('utf-8')
            status = f['status']

            try:
                patch = f['patch'].encode('utf-8')
            except KeyError:
                continue

            diff.append(b'diff --git a/%s b/%s' % (filename, filename))

            if status == 'modified':
                old_sha = file_shas[filename]
                new_sha = f['sha'].encode('utf-8')
                diff.append(b'index %s..%s 100644' % (old_sha, new_sha))
                diff.append(b'--- a/%s' % filename)
                diff.append(b'+++ b/%s' % filename)
            elif status == 'added':
                new_sha = f['sha'].encode('utf-8')

                diff.append(b'new file mode 100644')
                diff.append(b'index %s..%s' % (b'0' * 40, new_sha))
                diff.append(b'--- /dev/null')
                diff.append(b'+++ b/%s' % filename)
            elif status == 'removed':
                old_sha = file_shas[filename]

                diff.append(b'deleted file mode 100644')
                diff.append(b'index %s..%s' % (old_sha, b'0' * 40))
                diff.append(b'--- a/%s' % filename)
                diff.append(b'+++ /dev/null')
            elif status == 'renamed':
                old_filename = f['previous_filename'].encode('utf-8')
                old_sha = file_shas[old_filename]
                new_sha = f['sha'].encode('utf-8')

                diff.append(b'rename from %s' % old_filename)
                diff.append(b'rename to %s' % filename)
                diff.append(b'index %s..%s' % (old_sha, new_sha))
                diff.append(b'--- a/%s' % old_filename)
                diff.append(b'+++ b/%s' % filename)

            diff.append(patch)

        if diff and not diff[-1].endswith(b'\n'):
            # Make sure there's a trailing newline.
            diff.append(b'')

        diff = b'\n'.join(diff)

        return Commit(author_name=author_name,
                      id=revision,
                      date=date,
                      message=message,
                      parent=parent_revision,
                      diff=diff)

    def get_remote_repositories(self, owner=None, owner_type='user',
                                filter_type=None, start=None, per_page=None):
        """Return a list of remote repositories matching the given criteria.

        This will look up each remote repository on GitHub that the given
        owner either owns or is a member of.

        If the plan is an organization plan, then `owner` is expected to be
        an organization name, and the resulting repositories with be ones
        either owned by that organization or that the organization is a member
        of, and can be accessed by the authenticated user.

        If the plan is a public or private plan, and `owner` is the current
        user, then that user's public and private repositories or ones
        they're a member of will be returned.

        Otherwise, `owner` is assumed to be another GitHub user, and their
        accessible repositories that they own or are a member of will be
        returned.

        `owner` defaults to the linked account's username, and `plan`
        defaults to 'public'.
        """
        if owner is None and owner_type == 'user':
            owner = self.account.username

        assert owner

        url = self.get_api_url(self.account.hosting_url)
        paginator = self.client.api_get_remote_repositories(
            url, owner, owner_type, filter_type, start, per_page)

        return ProxyPaginator(
            paginator,
            normalize_page_data_func=lambda page_data: [
                RemoteRepository(
                    self,
                    repository_id='%s/%s' % (repo['owner']['login'],
                                             repo['name']),
                    name=repo['name'],
                    owner=repo['owner']['login'],
                    scm_type='Git',
                    path=repo['clone_url'],
                    mirror_path=repo['mirror_url'],
                    extra_data=repo)
                for repo in page_data
            ])

    def get_remote_repository(self, repository_id):
        """Get the remote repository for the ID.

        The ID is expected to be an ID returned from get_remote_repositories(),
        in the form of "owner/repo_id".

        If the repository is not found, ObjectDoesNotExist will be raised.
        """
        parts = repository_id.split('/')
        repo = None

        if len(parts) == 2:
            repo = self.client.api_get_remote_repository(
                self.get_api_url(self.account.hosting_url),
                *parts)

        if not repo:
            raise ObjectDoesNotExist

        return RemoteRepository(self,
                                repository_id=repository_id,
                                name=repo['name'],
                                owner=repo['owner']['login'],
                                scm_type='Git',
                                path=repo['clone_url'],
                                mirror_path=repo['mirror_url'],
                                extra_data=repo)

    def get_bug_info_uncached(
        self,
        repository: Repository,
        bug_id: str,
    ) -> BugInfo:
        """Return the information for the specified bug.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository object.

            bug_id (str):
                The ID of the bug to fetch.

        Returns:
            reviewboard.hostingsvcs.base.bug_tracker.BugInfo:
            Information about the bug.
        """
        result: BugInfo = {
            'summary': '',
            'description': '',
            'status': '',
        }

        repo_api_url = self._get_repo_api_url(repository)

        try:
            issue = self.client.api_get_issue(repo_api_url, bug_id)
            result = {
                'summary': issue['title'],
                'description': issue['body'],
                'status': issue['state'],
            }
        except Exception:
            # Errors in fetching are already logged in api_get_issue
            pass

        return result

    def get_repository_hook_instructions(self, request, repository):
        """Returns instructions for setting up incoming webhooks."""
        plan = repository.extra_data['repository_plan']
        add_webhook_url = urljoin(
            self.account.hosting_url or 'https://github.com/',
            '%s/%s/settings/hooks/new'
            % (self._get_repository_owner_raw(plan, repository.extra_data),
               self._get_repository_name_raw(plan, repository.extra_data)))

        webhook_endpoint_url = build_server_url(local_site_reverse(
            'github-hooks-close-submitted',
            local_site=repository.local_site,
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': repository.hosting_account.service_name,
            }))

        example_id = 123
        example_url = build_server_url(local_site_reverse(
            'review-request-detail',
            local_site=repository.local_site,
            kwargs={
                'review_request_id': example_id,
            }))

        return render_to_string(
            template_name='hostingsvcs/github/repo_hook_instructions.html',
            request=request,
            context={
                'example_id': example_id,
                'example_url': example_url,
                'repository': repository,
                'server_url': get_server_url(),
                'add_webhook_url': add_webhook_url,
                'webhook_endpoint_url': webhook_endpoint_url,
                'hook_uuid': repository.get_or_create_hooks_uuid(),
            })

    def _get_repo_api_url(self, repository):
        plan = repository.extra_data['repository_plan']

        return self._get_repo_api_url_raw(
            self._get_repository_owner_raw(plan, repository.extra_data),
            self._get_repository_name_raw(plan, repository.extra_data))

    def _get_repo_api_url_raw(self, owner, repo_name):
        return '%srepos/%s/%s' % (self.get_api_url(self.account.hosting_url),
                                  owner, repo_name)

    def _get_repository_owner_raw(self, plan, extra_data):
        if plan in ('public', 'private'):
            return self.account.username
        elif plan in ('public-org', 'private-org'):
            return self.get_plan_field(plan, extra_data, 'name')
        else:
            raise InvalidPlanError(plan)

    def _get_repository_name_raw(self, plan, extra_data):
        return self.get_plan_field(plan, extra_data, 'repo_name')



"""Update WebHookTarget.extra_data to allow NULL.

Version Added:
    2.5
"""

from django_evolution.mutations import ChangeField


MUTATIONS = [
    ChangeField('WebHookTarget', 'extra_data', initial=None, null=True)
]



"""Root-level resource for querying reviews.

Version Added:
    5.0
"""

from django.contrib.auth.models import User
from django.db.models import Q
from djblets.util.decorators import augment_method_from
from djblets.webapi.decorators import webapi_request_fields
from djblets.webapi.fields import (BooleanFieldType,
                                   DateTimeFieldType,
                                   StringFieldType)

from reviewboard.reviews.models import Group, Review
from reviewboard.scmtools.models import Repository
from reviewboard.webapi.decorators import webapi_check_local_site
from reviewboard.webapi.resources.base_review import BaseReviewResource


class RootReviewResource(BaseReviewResource):
    """Provides information on reviews.

    This is a top level endpoint that allows you to list and query all
    reviews in the system.

    Version Changed:
        5.0.4:
        This was formally named ``reviews`` on the root resource, but had to be
        renamed to ``all_reviews`` in order to work around a conflict with URI
        templates.

    Version Added:
        5.0
    """

    added_in = '5.0'
    allowed_methods = ('GET',)
    model = Review
    link_name = 'all_reviews'
    uri_template_name = 'all_review'
    uri_object_key = None

    @webapi_check_local_site
    def get_queryset(self, request, is_list=False, *args, **kwargs):
        """Return a queryset for Review models.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            is_list (bool, unused):
                Whether or not the queryset is for listing results.

            *args (tuple):
                Additional positional arguments.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            django.db.models.query.QuerySet:
            A queryset for Review models.
        """
        q = Q(**self.get_base_reply_to_field(*args, **kwargs))
        public = None

        if 'public' in request.GET:
            public = request.GET.get('public') in ('1', 'true', 'True')

        if 'user' in request.GET:
            user = list((
                User.objects
                .filter(username=request.GET.get('user'))
                .values_list('pk', flat=True)
            ))

            if user:
                q &= Q(user=user[0])
            else:
                return self.model.objects.none()

        if 'repository' in request.GET:
            repository = list((
                Repository.objects
                .filter(name=request.GET.get('repository'))
                .values_list('pk', flat=True)
            ))

            if repository:
                q &= Q(review_request__repository=repository[0])
            else:
                return self.model.objects.none()

        if 'last-updated-from' in request.GET:
            q &= Q(timestamp__gte=request.GET.get('last-updated-from'))

        if 'last-updated-to' in request.GET:
            q &= Q(timestamp__lt=request.GET.get('last-updated-to'))

        if 'review-group' in request.GET:
            users = list((
                Group.objects
                .filter(name=request.GET.get('review-group'))
                .values_list('users')
            ))
            q &= Q(user__in=users)

        return self.model.objects.accessible(request.user,
                                             extra_query=q,
                                             public=public,
                                             local_site=request.local_site)

    def get_base_reply_to_field(self, *args, **kwargs):
        """Return the base reply-to field for this resource.

        Args:
            *args (tuple, unused):
                Additional positional arguments.

            **kwargs (dict, unused):
                Additional keyword arguments.

        Returns:
            dict:
            A dictionary containing the base reply-to field.
        """
        return {
            'base_reply_to__isnull': True,
        }

    @webapi_request_fields(
        optional=dict({
            'last-updated-from': {
                'type': DateTimeFieldType,
                'description': "The earliest date/time the review could "
                               "be last updated. This is compared against "
                               "the review's ``timestamp`` field. This "
                               "must be a valid :term:`date/time format`.",
            },
            'last-updated-to': {
                'type': DateTimeFieldType,
                'description': "The date/time that the reviews must be "
                               "last updated before. This is compared against "
                               "the review\'s ``timestamp`` field. This "
                               "must be a valid :term:`date/time format`.",
            },
            'public': {
                'type': BooleanFieldType,
                'description': 'Whether to filter for public (published) '
                               'reviews. If not set, both published and '
                               'unpublished reviews will be included.',
            },
            'repository': {
                'type': StringFieldType,
                'description': 'The repository name that the review requests '
                               'of the reviews must be part of.',
            },
            'review-group': {
                'type': StringFieldType,
                'description': 'The group name of users that the reviews '
                               'must be owned by.',
            },
            'user': {
                'type': StringFieldType,
                'description': 'The username of the user that the reviews '
                               'must be owned by.',
            },
        }, **BaseReviewResource.get_list.optional_fields),
        required=BaseReviewResource.get_list.required_fields,
        allow_unknown=True
    )
    @augment_method_from(BaseReviewResource)
    def get_list(self, *args, **kwargs):
        """Return the list of reviews."""
        pass


root_review_resource = RootReviewResource()



"""Unit tests for reviewboard.accounts.service_accounts.ServiceAccountRegistry.

Version Added:
    8.1
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from django.contrib.auth.models import User
from django.db.models import Q
from django_assert_queries import assert_queries

from reviewboard.accounts.models import Profile
from reviewboard.accounts.service_accounts import (ServiceAccount,
                                                   ServiceAccountRegistry)
from reviewboard.testing import TestCase

if TYPE_CHECKING:
    from django_assert_queries import ExpectedQueries


class ServiceAccountRegistryTests(TestCase):
    """Unit tests for ServiceAccountRegistry.

    Version Added:
        8.1
    """

    def test_register_with_new_user(self) -> None:
        """Testing ServiceAccountRegistry.register with new user"""
        service_account = ServiceAccount(
            service_account_id='my-service-account',
        )

        registry = ServiceAccountRegistry()

        equeries: ExpectedQueries = [
            {
                'model': User,
                'select_related': {'profile'},
                'where': Q(username='my-service-account'),
            },
            {
                'model': User,
                'type': 'INSERT',
            },
            {
                'model': User,
                'type': 'UPDATE',
                'where': Q(pk=1),
            },
            {
                'model': Profile,
                'type': 'INSERT',
            },
        ]

        with assert_queries(equeries):
            registry.register(service_account)

        self.assertEqual(
            registry._by_username,
            {
                'my-service-account': service_account,
            })

    def test_register_with_existing_user(self) -> None:
        """Testing ServiceAccountRegistry.register with existing user"""
        service_account = ServiceAccount(
            service_account_id='my-service-account',
        )
        service_account.get_user()
        service_account._user = None

        registry = ServiceAccountRegistry()

        equeries: ExpectedQueries = [
            {
                'model': User,
                'select_related': {'profile'},
                'where': Q(username='my-service-account'),
            },
        ]

        with assert_queries(equeries):
            registry.register(service_account)

        self.assertEqual(
            registry._by_username,
            {
                'my-service-account': service_account,
            })

    def test_unregister(self) -> None:
        """Testing ServiceAccount.unregister"""
        service_account1 = ServiceAccount(
            service_account_id='my-service-account1',
        )
        service_account2 = ServiceAccount(
            service_account_id='my-service-account2',
        )

        registry = ServiceAccountRegistry()
        registry.register(service_account1)
        registry.register(service_account2)

        self.assertEqual(
            registry._by_username,
            {
                'my-service-account1': service_account1,
                'my-service-account2': service_account2,
            })

        registry.unregister(service_account2)

        self.assertEqual(
            registry._by_username,
            {
                'my-service-account1': service_account1,
            })

        registry.unregister(service_account1)

        self.assertEqual(registry._by_username, {})

    def test_get_for_service_account_id_found(self) -> None:
        """Testing ServiceAccountRegistry.get_for_service_account_id with
        service account found
        """
        service_account = ServiceAccount(
            service_account_id='my-service-account',
        )

        registry = ServiceAccountRegistry()
        registry.register(service_account)

        self.assertIs(
            registry.get_for_service_account_id('my-service-account'),
            service_account)

    def test_get_for_service_account_id_not_found(self) -> None:
        """Testing ServiceAccountRegistry.get_for_service_account_id with
        service account not found
        """
        registry = ServiceAccountRegistry()
        registry.register(ServiceAccount(
            service_account_id='my-service-account',
        ))

        self.assertIsNone(
            registry.get_for_service_account_id('other-service-account'))

    def test_get_for_username_found(self) -> None:
        """Testing ServiceAccountRegistry.get_for_username with service
        account found
        """
        service_account = ServiceAccount(
            service_account_id='my-service-account',
            preferred_username='my-service-user',
        )

        registry = ServiceAccountRegistry()
        registry.register(service_account)

        self.assertIs(registry.get_for_username('my-service-user'),
                      service_account)

    def test_get_for_username_not_found(self) -> None:
        """Testing ServiceAccountRegistry.get_for_username with service
        account not found
        """
        registry = ServiceAccountRegistry()
        registry.register(ServiceAccount(
            service_account_id='my-service-account',
            preferred_username='my-service-user',
        ))

        self.assertIsNone(registry.get_for_username('other-service-user'))



"""Unit tests for reviewboard.actions.renderers.SidebarActionGroupRenderer.

Version Added:
    8.0
"""

from __future__ import annotations

from django.template import Context
from django.utils.safestring import SafeString

from reviewboard.actions.base import (ActionPlacement,
                                      AttachmentPoint,
                                      BaseAction,
                                      BaseGroupAction)
from reviewboard.actions.renderers import SidebarActionGroupRenderer
from reviewboard.actions.tests.base import TestActionsRegistry
from reviewboard.testing import TestCase


class _MySidebarGroupAction(BaseGroupAction):
    action_id = 'test-sidebar-group'
    label = 'My Group'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.HEADER)
    ]


class _MySidebarItem1(BaseAction):
    action_id = 'test-sidebar-item1'
    label = 'Item 1'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.HEADER,
                        parent_id=_MySidebarGroupAction.action_id),
    ]


class _MySidebarSubGroup1Action(BaseGroupAction):
    action_id = 'test-sidebar-subgroup1'
    label = 'My Sub-Group 1'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.HEADER,
                        parent_id=_MySidebarGroupAction.action_id),
    ]


class _MySidebarItem2(BaseAction):
    action_id = 'test-sidebar-item2'
    label = 'Item 2'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.HEADER,
                        parent_id=_MySidebarSubGroup1Action.action_id),
    ]


class _MySidebarSubGroup2Action(BaseGroupAction):
    action_id = 'test-sidebar-subgroup2'
    label = 'My Sub-Group 2'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.HEADER,
                        parent_id=_MySidebarGroupAction.action_id),
    ]


class _MySidebarItem3(BaseAction):
    action_id = 'test-sidebar-item3'
    label = 'Item 3'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.HEADER,
                        parent_id=_MySidebarSubGroup2Action.action_id),
    ]


class SidebarActionGroupRendererTests(TestCase):
    """Unit tests for SidebarActionGroupRenderer.

    Version Added:
        8.0
    """

    def test_render(self) -> None:
        """Testing SidebarActionGroupRenderer.render"""
        group_action = _MySidebarGroupAction()
        subgroup1_action = _MySidebarSubGroup1Action()
        subgroup2_action = _MySidebarSubGroup2Action()
        item1_action = _MySidebarItem1()
        item2_action = _MySidebarItem2()
        item3_action = _MySidebarItem3()

        placement = group_action.get_placement(AttachmentPoint.HEADER)

        registry = TestActionsRegistry()
        registry.register(group_action)
        registry.register(item1_action)
        registry.register(subgroup1_action)
        registry.register(subgroup2_action)
        registry.register(item2_action)
        registry.register(item3_action)

        renderer = SidebarActionGroupRenderer(action=group_action,
                                              placement=placement)
        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        html = renderer.render(request=request,
                               context=context)

        self.assertIsInstance(html, SafeString)
        self.assertHTMLEqual(
            html,
            """
            <li class="rb-c-actions__action rb-c-sidebar__section"
                id="action-header-test-sidebar-group"
                aria-labelledby="test-sidebar-group__label"
                role="group">
             <header class="rb-c-sidebar__section-header"
                     id="test-sidebar-group__label">
              My Group
             </header>
             <ul class="rb-c-sidebar__items">
              <li class="rb-c-sidebar__nav-item"
                  id="action-header-test-sidebar-item1">
               <a class="rb-c-sidebar__item-label">
                Item 1
               </a>
              </li>
              <li class="rb-c-actions__action rb-c-sidebar__section"
                  id="action-header-test-sidebar-subgroup1"
                  aria-labelledby="test-sidebar-subgroup1__label"
                  role="group">
               <header class="rb-c-sidebar__section-header"
                       id="test-sidebar-subgroup1__label">
                My Sub-Group 1
               </header>
               <ul class="rb-c-sidebar__items">
                <li class="rb-c-sidebar__nav-item"
                    id="action-header-test-sidebar-item2">
                 <a class="rb-c-sidebar__item-label">
                  Item 2
                 </a>
                </li>
               </ul>
              </li>
              <li class="rb-c-actions__action rb-c-sidebar__section"
                  id="action-header-test-sidebar-subgroup2"
                  aria-labelledby="test-sidebar-subgroup2__label"
                  role="group">
               <header class="rb-c-sidebar__section-header"
                       id="test-sidebar-subgroup2__label">
                My Sub-Group 2
               </header>
               <ul class="rb-c-sidebar__items">
                <li class="rb-c-sidebar__nav-item"
                    id="action-header-test-sidebar-item3">
                 <a class="rb-c-sidebar__item-label">
                  Item 3
                 </a>
                </li>
               </ul>
              </li>
             </ul>
            </li>
            """)

    def test_render_js(self) -> None:
        """Testing SidebarActionGroupRenderer.render_js"""
        group_action = _MySidebarGroupAction()
        subgroup1_action = _MySidebarSubGroup1Action()
        subgroup2_action = _MySidebarSubGroup2Action()
        item1_action = _MySidebarItem1()
        item2_action = _MySidebarItem2()
        item3_action = _MySidebarItem3()
        placement = group_action.get_placement(AttachmentPoint.HEADER)

        registry = TestActionsRegistry()
        registry.register(group_action)
        registry.register(item1_action)
        registry.register(subgroup1_action)
        registry.register(subgroup2_action)
        registry.register(item2_action)
        registry.register(item3_action)

        renderer = SidebarActionGroupRenderer(action=group_action,
                                              placement=placement)
        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        js = renderer.render_js(request=request,
                                context=context)

        self.assertIsInstance(js, SafeString)
        self.assertHTMLEqual(
            js,
            """
            page.addActionView(new RB.Actions.ActionView({
                "attachmentPointID": "header",
                el: $('#action-header-test-sidebar-group'),
                model: page.getAction("test-sidebar-group"),
            }));
            """)



"""Unit tests for reviewboard.reviews.views.BatchOperationView."""

from __future__ import annotations

import json
from typing import List, TYPE_CHECKING

import kgb
from django_assert_queries.testing import assert_queries
from django.contrib.auth.models import User
from django.core import mail
from django.db.models import Q
from django.http.response import HttpResponse
from djblets.testing.decorators import add_fixtures

from reviewboard.accounts.models import (LocalSiteProfile,
                                         Profile,
                                         ReviewRequestVisit)
from reviewboard.changedescs.models import ChangeDescription
from reviewboard.notifications.models import WebHookTarget
from reviewboard.notifications.tests.mixins import EmailTestHelper
from reviewboard.reviews.models import (Group,
                                        Review,
                                        ReviewRequest,
                                        ReviewRequestDraft)
from reviewboard.reviews.testing.queries.review_requests import (
    get_review_requests_accessible_equeries,
)
from reviewboard.site.models import LocalSite
from reviewboard.site.urlresolvers import local_site_reverse
from reviewboard.testing import TestCase
from reviewboard.webapi.testing.queries import (
    get_webapi_request_start_equeries,
)

if TYPE_CHECKING:
    from django_assert_queries.query_comparator import ExpectedQueries


class BatchOperationViewTests(kgb.SpyAgency, EmailTestHelper, TestCase):
    """Unit tests for reviewboard.reviews.views.BatchOperationView."""

    fixtures = ['test_users']
    email_siteconfig_settings = {
        'mail_send_review_mail': True,
        'mail_default_from': 'noreply@example.com',
    }

    def setUp(self) -> None:
        """Set up the test case."""
        super().setUp()
        self.url = local_site_reverse('batch-operation')
        self.local_site_url = local_site_reverse(
            'batch-operation', local_site_name=self.local_site_name)

    def test_invalid(self) -> None:
        """Testing BatchOperationView with invalid data"""
        self.client.login(username='doc', password='doc')

        response = self.client.post(self.url, data={})
        self.assertResponse(response, 400, {
            'stat': 'fail',
            'error': 'Batch data was not found.',
        })

        response = self.client.post(self.url, data={
            'batch': 'aoeu',
        })
        self.assertResponse(response, 400, {
            'stat': 'fail',
            'error': 'Could not parse batch data: Expecting value: line '
                     '1 column 1 (char 0).',
        })

        response = self.client.post(self.url, data={
            'batch': json.dumps({}),
        })
        self.assertResponse(response, 400, {
            'stat': 'fail',
            'error': 'Unknown batch operation "None".',
        })

        response = self.client.post(self.url, data={
            'batch': json.dumps({
                'op': 'unknown',
            }),
        })
        self.assertResponse(response, 400, {
            'stat': 'fail',
            'error': 'Unknown batch operation "unknown".',
        })

    def test_archive(self) -> None:
        """Testing BatchOperationView archive op"""
        self.client.login(username='doc', password='doc')
        user = User.objects.get(username='doc')

        rr1, rr2, rr3 = self.create_many_review_requests(3, public=True)

        # Prime the caches.
        LocalSite.objects.has_local_sites()

        # 11 queries:
        #
        # 1-7. Review request list queries
        #   8. Try to fetch first review request visit
        #   9. Create first review request visit
        #  10. Try to fetch second review request visit
        #  11. Create second review request visit
        queries = self._get_review_request_list_queries(user, [rr1, rr2]) + [
            {
                'model': ReviewRequestVisit,
                'select_for_update': True,
                'where': Q(user=user, review_request=rr1),
            },
            {
                'model': ReviewRequestVisit,
                'type': 'INSERT',
            },
            {
                'model': ReviewRequestVisit,
                'select_for_update': True,
                'where': Q(user=user, review_request=rr2),
            },
            {
                'model': ReviewRequestVisit,
                'type': 'INSERT',
            },
        ]

        with assert_queries(queries, num_statements=19):
            response = self.client.post(self.url, data={
                'batch': json.dumps({
                    'op': 'archive',
                    'review_requests': [rr1.pk, rr2.pk],
                }),
            })
        self.assertEqual(response.status_code, 200)

        self.assertVisibility(rr1, 'doc', ReviewRequestVisit.ARCHIVED)
        self.assertVisibility(rr2, 'doc', ReviewRequestVisit.ARCHIVED)
        self.assertFalse(
            ReviewRequestVisit.objects.filter(
                review_request=rr3, user__username='doc')
            .exists())

    def test_archive_invalid(self) -> None:
        """Testing BatchOperationView archive op with invalid review
        requests
        """
        self.client.login(username='doc', password='doc')

        rr1, rr2, rr3 = self.create_many_review_requests(3, public=True)

        response = self.client.post(self.url, data={
            'batch': json.dumps({
                'op': 'archive',
                'review_requests': [rr1.pk, rr2.pk, 45],
            }),
        })
        self.assertResponse(response, 400, {
            'stat': 'fail',
            'error': 'The following review requests are not valid: [45].',
        })

    @add_fixtures(['test_site'])
    def test_archive_local_site(self) -> None:
        """Testing BatchOperationView archive op in a local site"""
        self.client.login(username='doc', password='doc')

        rr1, rr2, rr3 = self.create_many_review_requests(
            3, public=True, with_local_site=True)

        response = self.client.post(self.local_site_url, data={
            'batch': json.dumps({
                'op': 'archive',
                'review_requests': [rr1.display_id, rr2.display_id],
            }),
        })
        self.assertEqual(response.status_code, 200)

        self.assertVisibility(rr1, 'doc', ReviewRequestVisit.ARCHIVED)
        self.assertVisibility(rr2, 'doc', ReviewRequestVisit.ARCHIVED)
        self.assertFalse(
            ReviewRequestVisit.objects.filter(
                review_request=rr3, user__username='doc')
            .exists())

    @add_fixtures(['test_site'])
    def test_archive_local_site_invalid_ID(self) -> None:
        """Testing BatchOperationView archive op in a local site with
        invalid local_ids
        """
        self.client.login(username='doc', password='doc')

        rr1, rr2, rr3 = self.create_many_review_requests(
            3, public=True, with_local_site=True)

        response = self.client.post(self.local_site_url, data={
            'batch': json.dumps({
                'op': 'archive',
                'review_requests': [rr1.display_id,
                                    rr2.display_id,
                                    rr3.pk],
            }),
        })
        self.assertResponse(response, 400, {
            'stat': 'fail',
            'error': 'The following review requests are not valid for the '
                     'local site: [3].',
        })

    @add_fixtures(['test_site'])
    def test_archive_local_site_no_access(self) -> None:
        """Testing BatchOperationView archive op in a local site with
        no access
        """
        self.client.login(username='grumpy', password='grumpy')

        rr1, rr2, rr3 = self.create_many_review_requests(
            3, public=True, with_local_site=True)

        response = self.client.post(self.local_site_url, data={
            'batch': json.dumps({
                'op': 'archive',
                'review_requests': [rr1.display_id,
                                    rr2.display_id],
            }),
        })
        self.assertEqual(response.status_code, 403)

    def test_mute(self) -> None:
        """Testing BatchOperationView mute op"""
        self.client.login(username='doc', password='doc')
        user = User.objects.get(username='doc')

        rr1, rr2, rr3 = self.create_many_review_requests(3, public=True)

        # Prime the caches.
        LocalSite.objects.has_local_sites()

        # 11 queries:
        #
        #  1-7. Review request list queries
        #    8. Try to fetch first review request visit
        #    9. Create first review request visit
        #   10. Try to fetch second review request visit
        #   11. Create second review request visit
        queries = self._get_review_request_list_queries(user, [rr1, rr2]) + [
            {
                'model': ReviewRequestVisit,
                'select_for_update': True,
                'where': Q(user=user, review_request=rr1),
            },
            {
                'model': ReviewRequestVisit,
                'type': 'INSERT',
            },
            {
                'model': ReviewRequestVisit,
                'select_for_update': True,
                'where': Q(user=user, review_request=rr2),
            },
            {
                'model': ReviewRequestVisit,
                'type': 'INSERT',
            },
        ]

        with assert_queries(queries, num_statements=19):
            response = self.client.post(self.url, data={
                'batch': json.dumps({
                    'op': 'mute',
                    'review_requests': [rr1.pk, rr2.pk],
                }),
            })

        self.assertEqual(response.status_code, 200)

        self.assertVisibility(rr1, 'doc', ReviewRequestVisit.MUTED)
        self.assertVisibility(rr2, 'doc', ReviewRequestVisit.MUTED)
        self.assertFalse(
            ReviewRequestVisit.objects.filter(
                review_request=rr3, user__username='doc')
            .exists())

    def test_mute_invalid(self) -> None:
        """Testing BatchOperationView mute op with invalid review requests"""
        self.client.login(username='doc', password='doc')

        rr1, rr2, rr3 = self.create_many_review_requests(3, public=True)

        response = self.client.post(self.url, data={
            'batch': json.dumps({
                'op': 'mute',
                'review_requests': [rr1.pk, rr2.pk, 45],
            }),
        })
        self.assertResponse(response, 400, {
            'stat': 'fail',
            'error': 'The following review requests are not valid: [45].',
        })

    @add_fixtures(['test_site'])
    def test_mute_local_site(self) -> None:
        """Testing BatchOperationView mute op in a local site"""
        self.client.login(username='doc', password='doc')

        rr1, rr2, rr3 = self.create_many_review_requests(
            3, public=True, with_local_site=True)

        response = self.client.post(self.local_site_url, data={
            'batch': json.dumps({
                'op': 'mute',
                'review_requests': [rr1.display_id, rr2.display_id],
            }),
        })
        self.assertEqual(response.status_code, 200)

        self.assertVisibility(rr1, 'doc', ReviewRequestVisit.MUTED)
        self.assertVisibility(rr2, 'doc', ReviewRequestVisit.MUTED)
        self.assertFalse(
            ReviewRequestVisit.objects.filter(
                review_request=rr3, user__username='doc')
            .exists())

    @add_fixtures(['test_site'])
    def test_mute_local_site_invalid_ID(self) -> None:
        """Testing BatchOperationView mute op in a local site with invalid
        local_ids
        """
        self.client.login(username='doc', password='doc')

        rr1, rr2, rr3 = self.create_many_review_requests(
            3, public=True, with_local_site=True)

        response = self.client.post(self.local_site_url, data={
            'batch': json.dumps({
                'op': 'mute',
                'review_requests': [rr1.display_id, rr2.display_id, rr3.pk],
            }),
        })
        self.assertResponse(response, 400, {
            'stat': 'fail',
            'error': 'The following review requests are not valid for the '
                     'local site: [3].',
        })

    @add_fixtures(['test_site'])
    def test_mute_local_site_no_access(self) -> None:
        """Testing BatchOperationView mute op in a local site with no access"""
        self.client.login(username='grumpy', password='grumpy')

        rr1, rr2, rr3 = self.create_many_review_requests(
            3, public=True, with_local_site=True)

        response = self.client.post(self.local_site_url, data={
            'batch': json.dumps({
                'op': 'mute',
                'review_requests': [rr1.display_id, rr2.display_id],
            }),
        })
        self.assertEqual(response.status_code, 403)

    def test_unarchive(self) -> None:
        """Testing BatchOperationView unarchive op"""
        self.client.login(username='doc', password='doc')
        user = User.objects.get(username='doc')

        rr1, rr2, rr3 = self.create_many_review_requests(3, public=True)
        ReviewRequestVisit.objects.update_visibility(
            rr1, user, ReviewRequestVisit.ARCHIVED)
        ReviewRequestVisit.objects.update_visibility(
            rr2, user, ReviewRequestVisit.MUTED)
        ReviewRequestVisit.objects.update_visibility(
            rr3, user, ReviewRequestVisit.ARCHIVED)

        # Prime the caches.
        LocalSite.objects.has_local_sites()

        # 11 queries:
        #
        # 1-7. Review request list queries
        #   8. Try to fetch first review request visit
        #   9. Create first review request visit
        #  10. Try to fetch second review request visit
        #  11. Create second review request visit
        queries = self._get_review_request_list_queries(user, [rr1, rr2]) + [
            {
                'model': ReviewRequestVisit,
                'select_for_update': True,
                'where': Q(user=user, review_request=rr1),
            },
            {
                'model': ReviewRequestVisit,
                'type': 'UPDATE',
                'where': Q(pk=1),
            },
            {
                'model': ReviewRequestVisit,
                'select_for_update': True,
                'where': Q(user=user, review_request=rr2),
            },
            {
                'model': ReviewRequestVisit,
                'type': 'UPDATE',
                'where': Q(pk=2),
            },
        ]

        with assert_queries(queries, num_statements=15):
            response = self.client.post(self.url, data={
                'batch': json.dumps({
                    'op': 'unarchive',
                    'review_requests': [rr1.pk, rr2.pk],
                }),
            })
        self.assertEqual(response.status_code, 200)

        self.assertVisibility(rr1, 'doc', ReviewRequestVisit.VISIBLE)
        self.assertVisibility(rr2, 'doc', ReviewRequestVisit.VISIBLE)
        self.assertVisibility(rr3, 'doc', ReviewRequestVisit.ARCHIVED)

    def test_close(self) -> None:
        """Testing BatchOperationView close op"""
        self.client.login(username='doc', password='doc')
        user = User.objects.get(username='doc')
        profile = user.get_profile()
        site_profile = user.get_site_profile(None)

        rr1, rr2, rr3 = self.create_many_review_requests(3, public=True)

        # 39 queries:
        #
        #   1-7. Review request list queries
        #     8. Fetch user
        #     9. Fetch user
        # 10-25. Close first review request
        # 26-39. Close second review request
        queries = (  # noqa
            self._get_review_request_list_queries(user, [rr1, rr2]) +
            [
                {
                    'model': User,
                    'where': Q(id=user.pk),
                },
                {
                    'model': User,
                    'where': Q(id=user.pk),
                },
            ] +
            self._get_review_request_close_queries(
                user, profile, site_profile, rr1) +
            self._get_review_request_close_queries(
                user, profile, site_profile, rr2)
        )

        # Prime the caches.
        LocalSite.objects.has_local_sites()

        # We can't currently use this because QuerySet.update() is adding a
        # subquery that we can't test against.
        # with assert_queries(queries, num_statements=39):
        with self.assertNumQueries(39):
            response = self.client.post(self.url, data={
                'batch': json.dumps({
                    'op': 'close',
                    'review_requests': [rr1.pk, rr2.pk],
                }),
            })

        self.assertEqual(response.status_code, 200)

        rr1.refresh_from_db()
        rr2.refresh_from_db()
        rr3.refresh_from_db()

        self.assertEqual(rr1.status, ReviewRequest.SUBMITTED)
        self.assertEqual(rr2.status, ReviewRequest.SUBMITTED)
        self.assertEqual(rr3.status, ReviewRequest.PENDING_REVIEW)

    def test_close_invalid(self) -> None:
        """Testing BatchOperationView close op with invalid review requests"""
        self.client.login(username='doc', password='doc')

        rr1, rr2, rr3 = self.create_many_review_requests(3, public=True)

        response = self.client.post(self.url, data={
            'batch': json.dumps({
                'op': 'close',
                'review_requests': [rr1.pk, rr2.pk, 45],
            }),
        })
        self.assertResponse(response, 400, {
            'stat': 'fail',
            'error': 'The following review requests are not valid: [45].',
        })

    def test_close_permissions(self) -> None:
        """Testing BatchOperationView close op with non-owner user"""
        self.client.login(username='grumpy', password='grumpy')

        rr1, rr2, rr3 = self.create_many_review_requests(3, public=True)

        response = self.client.post(self.url, data={
            'batch': json.dumps({
                'op': 'close',
                'review_requests': [rr1.pk, rr2.pk],
            }),
        })
        self.assertResponse(response, 403, {
            'stat': 'fail',
            'error': 'User does not have permission to close review '
                     'request 1.',
        })

    @add_fixtures(['test_site'])
    def test_close_local_site(self) -> None:
        """Testing BatchOperationView close op in a local site"""
        self.client.login(username='doc', password='doc')

        rr1, rr2, rr3 = self.create_many_review_requests(
            3, public=True, with_local_site=True)

        response = self.client.post(self.local_site_url, data={
            'batch': json.dumps({
                'op': 'close',
                'review_requests': [rr1.display_id, rr2.display_id],
            }),
        })
        self.assertEqual(response.status_code, 200)

        rr1.refresh_from_db()
        rr2.refresh_from_db()
        rr3.refresh_from_db()

        self.assertEqual(rr1.status, ReviewRequest.SUBMITTED)
        self.assertEqual(rr2.status, ReviewRequest.SUBMITTED)
        self.assertEqual(rr3.status, ReviewRequest.PENDING_REVIEW)

    @add_fixtures(['test_site'])
    def test_close_local_site_invalid_ID(self) -> None:
        """Testing BatchOperationView close op in a local site with
        invalid local_ids
        """
        self.client.login(username='doc', password='doc')

        rr1, rr2, rr3 = self.create_many_review_requests(
            3, public=True, with_local_site=True)

        response = self.client.post(self.local_site_url, data={
            'batch': json.dumps({
                'op': 'close',
                'review_requests': [rr1.display_id,
                                    rr2.display_id,
                                    rr3.pk],
            }),
        })
        self.assertResponse(response, 400, {
            'stat': 'fail',
            'error': 'The following review requests are not valid for the '
                     'local site: [3].'
        })

        rr1.refresh_from_db()
        rr2.refresh_from_db()
        rr3.refresh_from_db()

        self.assertEqual(rr1.status, ReviewRequest.PENDING_REVIEW)
        self.assertEqual(rr2.status, ReviewRequest.PENDING_REVIEW)
        self.assertEqual(rr3.status, ReviewRequest.PENDING_REVIEW)

    def test_discard(self) -> None:
        """Testing BatchOperationView discard op"""
        self.client.login(username='doc', password='doc')
        user = User.objects.get(username='doc')
        profile = user.get_profile()
        site_profile = user.get_site_profile(None)

        rr1, rr2, rr3 = self.create_many_review_requests(3, public=True)

        # 39 queries:
        #
        #   1-7. Review request list queries
        #     8. Fetch user
        #     9. Fetch user
        # 10-25. Close first review request
        # 26-39. Close second review request
        queries = (  # noqa
            self._get_review_request_list_queries(user, [rr1, rr2]) +
            [
                {
                    'model': User,
                    'where': Q(id=user.pk),
                },
                {
                    'model': User,
                    'where': Q(id=user.pk),
                },
            ] +
            self._get_review_request_close_queries(
                user, profile, site_profile, rr1) +
            self._get_review_request_close_queries(
                user, profile, site_profile, rr2)
        )

        # Prime the caches.
        LocalSite.objects.has_local_sites()

        # We can't currently use this because QuerySet.update() is adding a
        # subquery that we can't test against.
        # with assert_queries(queries, num_statements=39):
        with self.assertNumQueries(39):
            response = self.client.post(self.url, data={
                'batch': json.dumps({
                    'op': 'discard',
                    'review_requests': [rr1.pk, rr2.pk],
                }),
            })

        self.assertEqual(response.status_code, 200)

        rr1.refresh_from_db()
        rr2.refresh_from_db()
        rr3.refresh_from_db()

        self.assertEqual(rr1.status, ReviewRequest.DISCARDED)
        self.assertEqual(rr2.status, ReviewRequest.DISCARDED)
        self.assertEqual(rr3.status, ReviewRequest.PENDING_REVIEW)

    def test_discard_invalid(self) -> None:
        """Testing BatchOperationView discard op with invalid review
        requests
        """
        self.client.login(username='doc', password='doc')

        rr1, rr2, rr3 = self.create_many_review_requests(3, public=True)

        response = self.client.post(self.url, data={
            'batch': json.dumps({
                'op': 'discard',
                'review_requests': [rr1.pk, rr2.pk, 45],
            }),
        })
        self.assertResponse(response, 400, {
            'stat': 'fail',
            'error': 'The following review requests are not valid: [45].',
        })

    def test_discard_permissions(self) -> None:
        """Testing BatchOperationView discard op with non-owner user"""
        self.client.login(username='grumpy', password='grumpy')

        rr1, rr2, rr3 = self.create_many_review_requests(3, public=True)

        response = self.client.post(self.url, data={
            'batch': json.dumps({
                'op': 'discard',
                'review_requests': [rr1.pk, rr2.pk],
            }),
        })
        self.assertResponse(response, 403, {
            'stat': 'fail',
            'error': 'User does not have permission to close review '
                     'request 1.',
        })

    @add_fixtures(['test_site'])
    def test_discard_local_site(self) -> None:
        """Testing BatchOperationView discard op in a local site"""
        self.client.login(username='doc', password='doc')

        rr1, rr2, rr3 = self.create_many_review_requests(
            3, public=True, with_local_site=True)

        response = self.client.post(self.local_site_url, data={
            'batch': json.dumps({
                'op': 'discard',
                'review_requests': [rr1.display_id,
                                    rr2.display_id],
            }),
        })
        self.assertEqual(response.status_code, 200)

        rr1.refresh_from_db()
        rr2.refresh_from_db()
        rr3.refresh_from_db()

        self.assertEqual(rr1.status, ReviewRequest.DISCARDED)
        self.assertEqual(rr2.status, ReviewRequest.DISCARDED)
        self.assertEqual(rr3.status, ReviewRequest.PENDING_REVIEW)

    @add_fixtures(['test_site'])
    def test_discard_local_site_invalid_ID(self) -> None:
        """Testing BatchOperationView discard op in a local site with
        invalid local_ids
        """
        self.client.login(username='doc', password='doc')

        rr1, rr2, rr3 = self.create_many_review_requests(
            3, public=True, with_local_site=True)

        response = self.client.post(self.local_site_url, data={
            'batch': json.dumps({
                'op': 'discard',
                'review_requests': [rr1.display_id,
                                    rr2.display_id,
                                    rr3.pk],
            }),
        })
        self.assertResponse(response, 400, {
            'stat': 'fail',
            'error': 'The following review requests are not valid for the '
                     'local site: [3].',
        })

        rr1.refresh_from_db()
        rr2.refresh_from_db()
        rr3.refresh_from_db()

        self.assertEqual(rr1.status, ReviewRequest.PENDING_REVIEW)
        self.assertEqual(rr2.status, ReviewRequest.PENDING_REVIEW)
        self.assertEqual(rr3.status, ReviewRequest.PENDING_REVIEW)

    def test_publish(self) -> None:
        """Testing BatchOperationView publish op with multiple objects"""
        self.client.login(username='doc', password='doc')
        grumpy = User.objects.get(username='grumpy')

        rr1 = self.create_review_request(publish=True, target_people=[grumpy])
        rr2 = self.create_review_request(target_people=[grumpy])
        self.create_review_request_draft(rr2)
        r1 = self.create_review(rr1, user='doc')
        r2 = self.create_review(rr1, user='doc')

        response = self.client.post(self.url, data={
            'batch': json.dumps({
                'op': 'publish',
                'review_requests': [rr2.pk],
                'reviews': [r1.pk, r2.pk],
            }),
        })
        self.assertEqual(response.status_code, 200)

        rr2.refresh_from_db()
        r1.refresh_from_db()
        r2.refresh_from_db()

        self.assertTrue(rr2.public)
        self.assertTrue(r1.public)
        self.assertTrue(r2.public)

    def test_publish_with_draft(self) -> None:
        """Testing BatchOperationView publish op with review request draft
        update
        """
        self.client.login(username='doc', password='doc')
        grumpy = User.objects.get(username='grumpy')

        rr = self.create_review_request(publish=True, target_people=[grumpy])
        draft = self.create_review_request_draft(rr)
        draft.summary = 'Updated Summary'
        draft.save(update_fields=('summary',))

        response = self.client.post(self.url, data={
            'batch': json.dumps({
                'op': 'publish',
                'review_requests': [rr.pk],
            }),
        })
        self.assertEqual(response.status_code, 200)

        rr.refresh_from_db()

        self.assertTrue(rr.public)
        self.assertEqual(rr.summary, 'Updated Summary')

    def test_publish_other_draft_as_admin(self) -> None:
        """Testing BatchOperationView publish op with an admin publishing
        a draft on another user's review request
        """
        self.client.login(username='admin', password='admin')
        grumpy = User.objects.get(username='grumpy')

        rr = self.create_review_request(publish=True,
                                        target_people=[grumpy])
        draft = self.create_review_request_draft(rr)
        draft.summary = 'Updated Summary'
        draft.save(update_fields=('summary',))

        response = self.client.post(self.url, data={
            'batch': json.dumps({
                'op': 'publish',
                'review_requests': [rr.pk],
            }),
        })
        self.assertEqual(response.status_code, 200)

        rr.refresh_from_db()

        self.assertTrue(rr.public)
        self.assertEqual(rr.summary, 'Updated Summary')

    def test_publish_permissions(self) -> None:
        """Testing BatchOperationView publish op with non-owner user"""
        self.client.login(username='grumpy', password='grumpy')
        grumpy = User.objects.get(username='grumpy')

        rr = self.create_review_request(publish=True, target_people=[grumpy])
        draft = self.create_review_request_draft(rr)
        draft.summary = 'Updated Summary'
        draft.save(update_fields=('summary',))

        response = self.client.post(self.url, data={
            'batch': json.dumps({
                'op': 'publish',
                'review_requests': [rr.pk],
            }),
        })
        self.assertResponse(response, 403, {
            'stat': 'fail',
            'error': 'User does not have permission to publish review '
                     'request 1.',
        })

    def test_publish_with_invalid_review_requests(self) -> None:
        """Testing BatchOperationView publish op with invalid review requests
        """
        self.client.login(username='doc', password='doc')
        response = self.client.post(self.url, data={
            'batch': json.dumps({
                'op': 'publish',
                'review_requests': [45],
            })
        })
        self.assertResponse(response, 400, {
            'stat': 'fail',
            'error': 'The following review requests are not valid: [45].',
        })

    def test_publish_with_invalid_reviews(self) -> None:
        """Testing BatchOperationView publish op with invalid reviews"""
        self.client.login(username='doc', password='doc')

        response = self.client.post(self.url, data={
            'batch': json.dumps({
                'op': 'publish',
                'review_requests': [],
                'reviews': [23],
            })
        })
        self.assertResponse(response, 400, {
            'stat': 'fail',
            'error': 'The following reviews are not valid: [23].',
        })

    def test_publish_with_publish_error(self) -> None:
        """Testing BatchOperationView publish op with publish error"""
        self.client.login(username='doc', password='doc')

        rr = self.create_review_request(publish=True)
        draft = self.create_review_request_draft(rr)
        draft.summary = 'Updated Summary'
        draft.save(update_fields=('summary',))

        response = self.client.post(self.url, data={
            'batch': json.dumps({
                'op': 'publish',
                'review_requests': [rr.pk],
            }),
        })
        self.assertResponse(response, 500, {
            'stat': 'fail',
            'error': 'Failed to publish review request 1: '
                     'Error publishing: There must be at least one '
                     'reviewer before this review request can be '
                     'published.',
            'review_requests_published': 0,
            'review_requests_not_published': 1,
            'reviews_published': 0,
            'reviews_not_published': 0,
        })

        rr.refresh_from_db()
        self.assertEqual(rr.summary, 'Test Summary')

        draft = rr.get_draft()
        self.assertIsNotNone(draft)
        self.assertEqual(draft.summary, 'Updated Summary')

    def test_publish_review_requests_mixed_result(self) -> None:
        """Testing BatchOperationView publish op with review requests and mixed
        result
        """
        self.client.login(username='doc', password='doc')
        grumpy = User.objects.get(username='grumpy')

        rr1 = self.create_review_request(publish=True, target_people=[grumpy])
        draft = self.create_review_request_draft(rr1)
        draft.summary = 'Updated Summary'
        draft.save(update_fields=('summary',))

        rr2 = self.create_review_request(publish=True)
        draft = self.create_review_request_draft(rr2)
        draft.summary = 'Updated Summary'
        draft.save(update_fields=('summary',))

        response = self.client.post(self.url, data={
            'batch': json.dumps({
                'op': 'publish',
                'review_requests': [rr1.pk, rr2.pk],
            }),
        })
        self.assertResponse(response, 500, {
            'stat': 'mixed',
            'error': 'Failed to publish review request 2: Error publishing: '
                     'There must be at least one reviewer before this review '
                     'request can be published.',
            'review_requests_not_published': 1,
            'review_requests_published': 1,
            'reviews_not_published': 0,
            'reviews_published': 0,
        })

    def test_publish_multiple_mixed_result(self) -> None:
        """Testing BatchOperationView publish op with multiple items and mixed
        result
        """
        self.client.login(username='doc', password='doc')
        grumpy = User.objects.get(username='grumpy')

        rr1 = self.create_review_request(publish=True, target_people=[grumpy])
        draft = self.create_review_request_draft(rr1)
        draft.summary = 'Updated Summary'
        draft.save(update_fields=('summary',))

        rr2 = self.create_review_request(publish=True, target_people=[grumpy])
        draft = self.create_review_request_draft(rr2)
        draft.summary = 'Updated Summary'
        draft.save(update_fields=('summary',))

        review1 = self.create_review(rr1, user='doc')
        review2 = self.create_review(rr2, user='doc')

        self.spy_on(
            Review.publish,
            owner=Review,
            op=kgb.SpyOpMatchInOrder([
                {
                },
                {
                    'op': kgb.SpyOpRaise(Exception('Cannot publish')),
                },
            ]))

        response = self.client.post(self.url, data={
            'batch': json.dumps({
                'op': 'publish',
                'review_requests': [rr1.pk, rr2.pk],
                'reviews': [review1.pk, review2.pk],
            }),
        })
        self.assertResponse(response, 500, {
            'stat': 'mixed',
            'error': 'Failed to publish review 2: Cannot publish',
            'review_requests_not_published': 0,
            'review_requests_published': 2,
            'reviews_not_published': 1,
            'reviews_published': 1,
        })

    def test_publish_trivial(self) -> None:
        """Testing BatchOperationView publish op with trivial flag"""
        self.client.login(username='doc', password='doc')
        grumpy = User.objects.get(username='grumpy')

        rr1 = self.create_review_request(publish=True, target_people=[grumpy])
        rr2 = self.create_review_request(target_people=[grumpy])
        self.create_review_request_draft(rr2)
        r1 = self.create_review(rr1, user='doc')
        r2 = self.create_review(rr1, user='doc')

        mail.outbox = []

        response = self.client.post(self.url, data={
            'batch': json.dumps({
                'op': 'publish',
                'review_requests': [rr2.pk],
                'reviews': [r1.pk, r2.pk],
                'trivial': True,
            }),
        })
        self.assertEqual(response.status_code, 200)

        self.assertEqual(len(mail.outbox), 0)

    def test_publish_notification_with_single_review_request(self) -> None:
        """Testing BatchOperationView publish op notification with a single
        review request
        """
        self.client.login(username='doc', password='doc')
        grumpy = User.objects.get(username='grumpy')

        rr1 = self.create_review_request(target_people=[grumpy])
        self.create_review_request_draft(rr1)

        mail.outbox = []

        response = self.client.post(self.url, data={
            'batch': json.dumps({
                'op': 'publish',
                'review_requests': [rr1.pk],
            }),
        })
        self.assertEqual(response.status_code, 200)

        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].subject,
                         'Review Request 1: Test Summary')

    def test_publish_review_request_and_review(self) -> None:
        """Testing BatchOperationView publish op with a new review request and
        review
        """
        self.client.login(username='doc', password='doc')
        grumpy = User.objects.get(username='grumpy')

        rr1 = self.create_review_request(target_people=[grumpy])
        self.create_review_request_draft(rr1)
        r1 = self.create_review(rr1, user='doc')

        response = self.client.post(self.url, data={
            'batch': json.dumps({
                'op': 'publish',
                'review_requests': [rr1.pk],
                'reviews': [r1.pk],
            }),
        })
        self.assertEqual(response.status_code, 200)

        rr1.refresh_from_db()
        r1.refresh_from_db()

        self.assertTrue(rr1.public)
        self.assertTrue(r1.public)

    def test_publish_review_on_unpublished_review_request(self) -> None:
        """Testing BatchOperationView publish op with a review on an
        unpublished review request
        """
        self.client.login(username='doc', password='doc')
        grumpy = User.objects.get(username='grumpy')

        rr1 = self.create_review_request(target_people=[grumpy])
        self.create_review_request_draft(rr1)
        r1 = self.create_review(rr1, user='doc')

        response = self.client.post(self.url, data={
            'batch': json.dumps({
                'op': 'publish',
                'review_requests': [],
                'reviews': [r1.pk],
            }),
        })
        self.assertResponse(response, 403, {
            'stat': 'fail',
            'error': 'This review cannot be published until the review '
                     'request is published.',
        })

        rr1.refresh_from_db()
        r1.refresh_from_db()

        self.assertFalse(rr1.public)
        self.assertFalse(r1.public)

    def test_publish_notifications(self) -> None:
        """Testing BatchOperationView publish op with multiple published items
        """
        self.client.login(username='doc', password='doc')
        grumpy = User.objects.get(username='grumpy')

        rr1, rr2, rr3 = self.create_many_review_requests(
            3, public=True, target_people=[grumpy])
        rr1.email_message_id = 'fake message id'
        rr1.save(update_fields=('email_message_id',))
        draft1 = self.create_review_request_draft(rr1)
        draft1.summary = 'Updated Summary'
        draft1.save(update_fields=('summary',))
        draft2 = self.create_review_request_draft(rr2)
        draft2.summary = 'Updated Summary 2'
        draft2.save(update_fields=('summary',))

        r1 = self.create_review(rr1, user='doc')
        r2 = self.create_review(rr2, user='doc')
        r3 = self.create_review(rr3, user='doc')

        mail.outbox = []

        response = self.client.post(self.url, data={
            'batch': json.dumps({
                'op': 'publish',
                'review_requests': [rr1.pk, rr2.pk],
                'reviews': [r1.pk, r2.pk, r3.pk],
            }),
        })
        self.assertEqual(response.status_code, 200)

        self.assertEqual(len(mail.outbox), 3)
        self.assertEqual(mail.outbox[0].subject,
                         'Re: Review Request 1: Updated Summary')
        self.assertEqual(mail.outbox[1].subject,
                         'Review Request 2: Updated Summary 2')
        self.assertEqual(mail.outbox[2].subject,
                         'Review Request 3: Test Summary 3')

    def test_publish_archive(self) -> None:
        """Testing BatchOperationView publish op with archive after publish"""
        self.client.login(username='doc', password='doc')
        grumpy = User.objects.get(username='grumpy')

        rr1 = self.create_review_request(publish=True, target_people=[grumpy])
        rr2 = self.create_review_request(target_people=[grumpy])
        self.create_review_request_draft(rr2)
        r1 = self.create_review(rr1, user='doc')
        r2 = self.create_review(rr1, user='doc')

        response = self.client.post(self.url, data={
            'batch': json.dumps({
                'op': 'publish',
                'review_requests': [rr2.pk],
                'reviews': [r1.pk, r2.pk],
                'archive': True,
            }),
        })
        self.assertEqual(response.status_code, 200)

        self.assertVisibility(rr1, 'doc', ReviewRequestVisit.ARCHIVED)
        self.assertVisibility(rr2, 'doc', ReviewRequestVisit.ARCHIVED)

    @add_fixtures(['test_scmtools'])
    def test_publish_with_acls(self) -> None:
        """Testing BatchOperationView when publishing a review on a review
        request that has multiple groups with ACLs
        """
        # Review Board 6.0 had a bug where the way the batch endpoint was
        # fetching reviews could return duplicates, which would then fail
        # length comparisons. This test case is an example of something that
        # triggered that bug before it was fixed.
        doc = User.objects.get(username='doc')
        grumpy = User.objects.get(username='grumpy')

        group1 = self.create_review_group(invite_only=True)
        group1.users.add(doc, grumpy)

        group2 = self.create_review_group(invite_only=True)
        group2.users.add(doc, grumpy)

        repository = self.create_repository(public=False)
        repository.review_groups.set([group1])

        review_request = self.create_review_request(
            publish=True,
            repository=repository,
            target_groups=[group1, group2])
        review = self.create_review(review_request, user='doc')

        self.client.login(username='doc', password='doc')
        response = self.client.post(self.url, data={
            'batch': json.dumps({
                'op': 'publish',
                'review_requests': [],
                'reviews': [review.pk],
            }),
        })
        self.assertEqual(response.status_code, 200)

        reply = self.create_reply(review, user='grumpy')
        self.client.login(username='grumpy', password='grumpy')
        response = self.client.post(self.url, data={
            'batch': json.dumps({
                'op': 'publish',
                'review_requests': [],
                'reviews': [reply.pk],
            }),
        })
        self.assertEqual(response.status_code, 200)

    def assertVisibility(
        self,
        review_request: ReviewRequest,
        user: str,
        visibility: str,
    ) -> None:
        """Assert a specific review request visibility state.

        Args:
            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request to check.

            user (str):
                The username of the user to check.

            visibility (str):
                The expected visibility state.
        """
        visit = ReviewRequestVisit.objects.get(
            review_request=review_request,
            user__username=user)
        self.assertEqual(visit.visibility, visibility)

    def assertResponse(
        self,
        response: HttpResponse,
        status_code: int,
        data: dict,
    ) -> None:
        """Assert that a response matches the expected result.

        Args:
            response (django.http.HttpResponse):
                The response from the test client.

            status_code (int):
                The expected HTTP status.

            data (dict):
                The expected data contained in the response after decoding
                JSON.
        """
        self.assertEqual(response.status_code, status_code)
        self.assertEqual(json.loads(response.content), data)

    def _get_review_request_list_queries(
        self,
        user: User,
        review_requests: List[ReviewRequest],
        has_local_sites_in_db: bool = False,
    ) -> ExpectedQueries:
        """Return queries for the initial review requests list fetch.

        Args:
            user (django.contrib.auth.models.User):
                The user making the request.

            review_requests (list of reviewboard.reviews.models.ReviewRequest):
                The review requests being passed to the batch endpoint.

            has_local_sites_in (bool, optional):
                Whether the test expects Local Sites to be present in the
                database.

        Returns:
            list:
            A list of query info appropriate for assert_queries.
        """
        equeries = get_webapi_request_start_equeries(user=user)
        equeries += get_review_requests_accessible_equeries(
            user=user,
            show_all_unpublished=True,
            status=None,
            extra_query=Q(pk__in=[
                review_request.pk
                for review_request in review_requests
            ]))

        return equeries

    def _get_review_request_close_queries(
        self,
        user: User,
        profile: Profile,
        site_profile: LocalSiteProfile,
        review_request: ReviewRequest,
    ) -> List[dict]:
        """Return queries associated with closing a review request.

        Args:
            user (django.contrib.auth.models.User):
                The user performing the close.

            profile (reviewboard.accounts.models.Profile):
                The user's profile.

            site_profile (reviewboard.accounts.models.LocalSiteProfile):
                The user's local site profile.

            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request being closed.

        Returns:
            list:
            A list of query info appropriate for assert_queries.
        """
        # 15 queries
        #
        #  1. Fetch review request draft
        #  2. Create ChangeDescription
        #  3. Link ChangeDescription
        #  4. Fetch user's Profile
        #  5. Fetch user's LocalSiteProfile
        #  6. Check review request status
        #  7. Update local site profile
        #  8. Refresh local site profile
        #  9. Fetch target groups
        # 10. Fetch target users
        # 11. Update group counters
        # 12. Update LocalSiteProfile counters
        # 13. Fetch ReviewRequestVisit
        # 14. Update ReviewRequestVisit
        # 15. Fetch WebHook targets
        return [
            {
                'model': ReviewRequestDraft,
                'where': Q(review_request=review_request),
            },
            {
                'model': ChangeDescription,
                'type': 'INSERT',
            },
            {
                'model': ReviewRequest.changedescs.through,
                'type': 'INSERT',
            },
            {
                'model': Profile,
                'where': Q(user=user),
            },
            {
                'model': LocalSiteProfile,
                'where': Q(local_site=None, profile=profile, user=user),
            },
            {
                'model': ReviewRequest,
                'where': Q(pk=review_request.pk),
                'only_fields': {'status', 'public'},
            },
            {
                'model': LocalSiteProfile,
                'type': 'UPDATE',
                'where': Q(pk=site_profile.pk),
            },
            {
                'model': LocalSiteProfile,
                'limit': 1,
                'values_select': ('pending_outgoing_request_count',),
                'where': Q(pk=site_profile.pk),
            },
            {
                'model': Group,
                'num_joins': 1,
                'tables': {
                    'reviews_reviewrequest_target_groups',
                    'reviews_group',
                },
                'where': Q(review_requests__id=review_request.pk),
                'values_select': ('pk',),
            },
            {
                'model': User,
                'num_joins': 1,
                'tables': {
                    'auth_user',
                    'reviews_reviewrequest_target_people',
                },
                'where': Q(directed_review_requests__id=review_request.pk),
                'values_select': ('pk',),
            },
            {
                'model': Group,
                'num_joins': 1,
                'tables': {
                    'reviews_reviewrequest_target_groups',
                    'reviews_group',
                },
                'type': 'UPDATE',
            },
            {
                'model': LocalSiteProfile,
                'type': 'UPDATE',
                'num_joins': 2,
                'tables': {
                    'accounts_profile',
                    'accounts_localsiteprofile',
                    'accounts_profile_starred_review_requests',
                },
            },
            {
                'model': ReviewRequestVisit,
                'where': Q(review_request=review_request),
            },
            {
                'model': ReviewRequest,
                'type': 'UPDATE',
                'where': Q(pk=review_request.pk),
            },
            {
                'model': WebHookTarget,
                'where': Q(Q(enabled=True, local_site=None) &
                           Q(Q(apply_to='A') | Q(apply_to='N'))),
            },
        ]



"""Admin-specific form widgets."""

from __future__ import annotations

import logging
from typing import Any, Dict, Iterable, Optional, TYPE_CHECKING

from django.contrib.auth.models import User
from django.template.loader import render_to_string
from django.utils.encoding import force_str
from django.utils.safestring import mark_safe
from django.forms.widgets import MultiWidget, Select, TextInput
from djblets.forms.widgets import (
    RelatedObjectWidget as DjbletsRelatedObjectWidget)
from pygments.lexers import get_all_lexers

from reviewboard.avatars import avatar_services
from reviewboard.reviews.models import Group
from reviewboard.scmtools.models import Repository

if TYPE_CHECKING:
    from django.forms.renderers import BaseRenderer
    from django.utils.safestring import SafeString


logger = logging.getLogger(__name__)


class RelatedObjectWidget(DjbletsRelatedObjectWidget):
    """A base class form widget that lets people select one or more objects.

    This is a base class. Extended classes must define their own render()
    method, to render their own widget with their own data.

    This should be used with relatedObjectSelectorView.es6.js, which extends
    a Backbone view to display data.
    """

    ######################
    # Instance variables #
    ######################

    #: The optional Local Site to bound the API requests to.
    local_site_name: Optional[str]

    def __init__(
        self,
        local_site_name: Optional[str] = None,
        multivalued: bool = True,
        **kwargs,
    ) -> None:
        """Initialize the widget.

        Args:
            local_site_name (str, optional):
                The optional Local Site to bound the API requests to.

            **kwargs (dict):
                Additional keyword arguments to pass to the parent.
        """
        super().__init__(multivalued=multivalued,
                         **kwargs)

        self.local_site_name = local_site_name


class RelatedUserWidget(RelatedObjectWidget):
    """A form widget to allow people to select one or more User relations.

    It's not unheard of to have a server with thousands or tens of thousands of
    registered users. In this case, the existing Django admin widgets fall down
    hard. The filtered select widgets can actually crash the webserver due to
    trying to pre-populate an enormous ``<select>`` element, and the raw ID
    widget is basically a write-only field.

    This field does much better, offering both the ability to see who's already
    in the list, as well as interactive search and filtering.
    """

    def render(self, name, value, attrs=None, renderer=None):
        """Render the widget.

        Args:
            name (unicode):
                The name of the field.

            value (list):
                The current value of the field.

            attrs (dict, optional):
                Attributes for the HTML element.

            renderer (django.forms.renderers.BaseRenderer, optional):
                The form renderer.

        Returns:
            django.utils.safestring.SafeText:
            The rendered HTML.
        """
        if value:
            if not self.multivalued:
                value = [value]

            value = [v for v in value if v]
            input_value = ','.join(force_str(v) for v in value)
            existing_users = (
                User.objects
                .filter(pk__in=value)
                .order_by('first_name', 'last_name', 'username')
            )
        else:
            input_value = None
            existing_users = []

        final_attrs = dict(self.attrs, **attrs)
        final_attrs['name'] = name

        input_html = super(RelatedUserWidget, self).render(
            name, input_value, attrs, renderer)

        use_avatars = avatar_services.avatars_enabled
        user_data = []

        for user in existing_users:
            data = {
                'fullname': user.get_full_name(),
                'id': user.pk,
                'username': user.username,
            }

            if use_avatars:
                try:
                    data['avatarHTML'] = (
                        avatar_services.for_user(user)
                        .render(request=None,
                                user=user,
                                size=20)
                    )
                except Exception as e:
                    logger.exception(
                        'Error rendering avatar for RelatedUserWidget: %s',
                        e)
                    data['avatarHTML'] = None

            user_data.append(data)

        return render_to_string(
            template_name='admin/related_user_widget.html',
            context={
                'input_html': mark_safe(input_html),
                'input_id': final_attrs['id'],
                'local_site_name': self.local_site_name,
                'multivalued': self.multivalued,
                'use_avatars': use_avatars,
                'users': user_data,
            })

    def value_from_datadict(self, data, files, name):
        """Unpack the field's value from a datadict.

        Args:
            data (dict):
                The form's data.

            files (dict):
                The form's files.

            name (unicode):
                The name of the field.

        Returns:
            list:
            The list of PKs of User objects.
        """
        value = data.get(name)

        if self.multivalued:
            if isinstance(value, list):
                return value
            elif isinstance(value, str):
                return [v for v in value.split(',') if v]
            else:
                return None
        elif value:
            return value
        else:
            return None


class RelatedRepositoryWidget(RelatedObjectWidget):
    """A form widget allowing people to select one or more Repository objects.

    This widget offers both the ability to see which repositories are already
    in the list, as well as interactive search and filtering.
    """

    ######################
    # Instance variables #
    ######################

    #: Whether to include accessible invisible repositories in the results.
    #:
    #: Version Added:
    #:     5.0.6
    show_invisible: bool

    def __init__(
        self,
        *args,
        show_invisible: bool = True,
        **kwargs,
    ) -> None:
        """Initialize the RelatedGroupWidget.

        Version Changed:
            5.0.6:
            Added the ``show_invisible`` argument.

        Args:
            show_invisible (bool, optional):
                Whether to include accessible invisible repositories in the
                results.

                This is the default.

                Version Added:
                    5.0.6

            *args (tuple):
                Positional arguments to pass to the handler.

            **kwargs (dict):
                Keyword arguments to pass to the handler.
        """
        super().__init__(*args, **kwargs)

        self.show_invisible = show_invisible

    def render(
        self,
        name: str,
        value: Any,
        attrs: Optional[Dict[str, Any]] = None,
        renderer: Optional[BaseRenderer] = None,
    ) -> SafeString:
        """Render the widget.

        Args:
            name (str):
                The name of the field.

            value (list):
                The current value of the field.

            attrs (dict, optional):
                Attributes for the HTML element.

            renderer (django.forms.renderers.BaseRenderer, optional):
                The form renderer.

        Returns:
            django.utils.safestring.SafeText:
            The rendered HTML.
        """
        existing_repos: Iterable[Repository]
        input_value: Optional[str]

        if value:
            if not self.multivalued:
                value = [value]

            value = [v for v in value if v]
            input_value = ','.join(force_str(v) for v in value)
            existing_repos = (
                Repository.objects
                .filter(pk__in=value)
                .order_by('name')
            )
        else:
            input_value = None
            existing_repos = []

        if attrs:
            final_attrs = dict(self.attrs, **attrs)
        else:
            final_attrs = self.attrs.copy()

        final_attrs['name'] = name

        input_html = super().render(name, input_value, attrs, renderer)

        repo_data = [
            {
                'id': repo.pk,
                'name': repo.name,
            }
            for repo in existing_repos
        ]

        js_view_data: Dict[str, Any] = {
            'initialOptions': repo_data,
            'multivalued': self.multivalued,
            'showInvisible': self.show_invisible,
        }

        if self.local_site_name:
            js_view_data['localSitePrefix'] = f's/{self.local_site_name}/'

        return render_to_string(
            template_name='admin/related_repo_widget.html',
            context={
                'input_html': mark_safe(input_html),
                'input_id': final_attrs['id'],
                'js_view_data': js_view_data,
            })

    def value_from_datadict(self, data, files, name):
        """Unpack the field's value from a datadict.

        Args:
            data (dict):
                The form's data.

            files (dict):
                The form's files.

            name (unicode):
                The name of the field.

        Returns:
            list:
            The list of IDs of
            :py:class:`~reviewboard.scmtools.models.Repository` objects.
        """
        value = data.get(name)

        if self.multivalued:
            if isinstance(value, list):
                return value
            elif isinstance(value, str):
                return [v for v in value.split(',') if v]
            else:
                return None
        elif value:
            return value
        else:
            return None


class RelatedGroupWidget(RelatedObjectWidget):
    """A form widget allowing people to select one or more Group objects.

    This widget offers both the ability to see which groups are already in the
    list, as well as interactive search and filtering.

    Version Changed:
        5.0.6:
        * Added an option for enabling specifying invisible review groups.
        * Added support for Python type hints.
    """

    ######################
    # Instance variables #
    ######################

    #: Whether or not to display review groups that are invite-only.
    invite_only: bool

    #: Whether or not to limit review groups to ones that are visible.
    #:
    #: Version Added:
    #:     5.0.6
    show_invisible: bool

    def __init__(
        self,
        invite_only: bool = False,
        *args,
        show_invisible: bool = True,
        **kwargs,
    ) -> None:
        """Initialize the RelatedGroupWidget.

        Version Changed:
            5.0.6:
            Added the ``show_invisible`` argument.

        Args:
            invite_only (bool, optional):
                Whether or not to limit results to accessible review groups
                that are invite-only.

            show_invisible (bool, optional):
                Whether to include accessible invisible review groups in the
                results.

                This is the default.

                Version Added:
                    5.0.6

            *args (tuple):
                Positional arguments to pass to the handler.

            **kwargs (dict):
                Keyword arguments to pass to the handler.
        """
        super().__init__(*args, **kwargs)

        self.invite_only = invite_only
        self.show_invisible = show_invisible

    def render(
        self,
        name: str,
        value: Any,
        attrs: Optional[Dict[str, Any]] = None,
        renderer: Optional[BaseRenderer] = None,
    ) -> SafeString:
        """Render the widget.

        Args:
            name (str):
                The name of the field.

            value (list):
                The current value of the field.

            attrs (dict, optional):
                Attributes for the HTML element.

            renderer (django.forms.renderers.BaseRenderer, optional):
                The form renderer.

        Returns:
            django.utils.safestring.SafeText:
            The rendered HTML.
        """
        existing_groups: Iterable[Group]
        input_value: Optional[str]

        if value:
            if not self.multivalued:
                value = [value]

            value = [v for v in value if v]
            input_value = ','.join(force_str(v) for v in value)
            existing_groups = (
                Group.objects
                .filter(pk__in=value)
                .order_by('name')
            )
        else:
            input_value = None
            existing_groups = []

        if attrs:
            final_attrs = dict(self.attrs, **attrs)
        else:
            final_attrs = self.attrs.copy()

        final_attrs['name'] = name

        input_html = super().render(name, input_value, attrs, renderer)

        group_data = []

        for group in existing_groups:
            data = {
                'name': group.name,
                'display_name': group.display_name,
                'id': group.pk,
            }

            group_data.append(data)

        js_view_data: Dict[str, Any] = {
            'initialOptions': group_data,
            'inviteOnly': self.invite_only,
            'multivalued': self.multivalued,
            'showInvisible': self.show_invisible,
        }

        if self.local_site_name:
            js_view_data['localSitePrefix'] = f's/{self.local_site_name}/'

        return render_to_string(
            template_name='admin/related_group_widget.html',
            context={
                'input_html': mark_safe(input_html),
                'input_id': final_attrs['id'],
                'js_view_data': js_view_data,
            })

    def value_from_datadict(self, data, files, name):
        """Unpack the field's value from a datadict.

        Args:
            data (dict):
                The form's data.

            files (dict):
                The form's files.

            name (unicode):
                The name of the field.

        Returns:
            list:
            The list of PKs of Group objects.
        """
        value = data.get(name)

        if self.multivalued:
            if isinstance(value, list):
                return value
            elif isinstance(value, str):
                return [v for v in value.split(',') if v]
            else:
                return None
        elif value:
            return value
        else:
            return None


class LexersMappingWidget(MultiWidget):
    """A form widget for mapping a string to a Pygments Lexer class.

    This widget displays a text input with a drop-down list of
    Pygments Lexer names next to it.

    Version Added:
        5.0
    """

    def __init__(self, attrs=None):
        """Initialize the LexersMappingWidget.

        Args:
            attrs (dict, optional):
                A dictionary containing HTML attributes to be set
                on the rendered widget.
        """
        lexer_choices = [(lex[0], lex[0]) for lex in get_all_lexers()]
        widgets = (
            TextInput(attrs=attrs),
            Select(attrs=attrs, choices=lexer_choices))
        super(LexersMappingWidget, self).__init__(widgets, attrs)

    def decompress(self, value):
        """Decompress the value into a list of values for each widget.

        Args:
            value (tuple of str):
                The value from the field. A tuple containing two strings,
                a key and a Pygments Lexer name.

        Returns:
            list of str:
            The list containing a key and lexer name for the widgets.
        """
        if value:
            return list(value)

        return [None, None]

    def value_from_datadict(self, data, files, name):
        """Unpack the field's value from a datadict.

        Args:
            data (dict):
                The form's data.

            files (dict):
                The form's files.

            name (str):
                The name of the field.

        Returns:
            tuple of str:
            The tuple containing a key and lexer name.
        """
        key_lexer = [
            widget.value_from_datadict(data, files, '%s_%s' % (name, i))
            for i, widget in enumerate(self.widgets)]

        if key_lexer:
            return tuple(key_lexer)
        else:
            return (None, None)



"""Evolution to set null=True on a review request's diffset history.

This only affects the stored signature, and should not impact the database
schema.

Version Added:
    7.0
"""

from django_evolution.mutations import ChangeField


MUTATIONS = [
    ChangeField('ReviewRequest', 'diffset_history', initial=None, null=True),
]



"""Review UI for the legacy Screenshot object."""

from __future__ import annotations

from typing import Optional, Sequence, TYPE_CHECKING

from reviewboard.reviews.models import (
    Screenshot,
    ScreenshotComment,
)
from reviewboard.reviews.ui.base import ReviewUI, SerializedCommentBlocks
from reviewboard.reviews.ui.image import SerializedRegionComment

if TYPE_CHECKING:
    from typelets.json import JSONDict

    from reviewboard.reviews.models import ReviewRequestDraft


class LegacyScreenshotReviewUI(ReviewUI[
    Screenshot,
    ScreenshotComment,
    SerializedRegionComment
]):
    """Review UI for the legacy Screenshot object."""

    name = 'Screenshot'
    model = Screenshot
    comment_model = ScreenshotComment

    js_model_class: str = 'RB.ScreenshotReviewable'
    js_view_class: str = 'RB.ImageReviewableView'

    def get_comments(self) -> Sequence[ScreenshotComment]:
        """Return all existing comments on the screenshot.

        Returns:
            list of reviewboard.reviews.screenshot_comment.ScreenshotComment:
            The list of comments for the page.
        """
        return self.obj.get_comments()

    def get_caption(
        self,
        draft: Optional[ReviewRequestDraft] = None,
    ) -> str:
        """Return the caption to show for the screenshot.

        Args:
            draft (reviewboard.reviews.models.review_request_draft.
                   ReviewRequestDraft, optional):
                The active review request draft for the user, if any.

        Returns:
            str:
            The caption for the screenshot.
        """
        assert isinstance(self.obj, Screenshot)
        if draft and self.obj.draft_caption:
            return self.obj.draft_caption

        return self.obj.caption

    def get_comment_link_text(
        self,
        comment: ScreenshotComment,
    ) -> str:
        """Return the text to link to a comment.

        Args:
            comment (reviewboard.reviews.models.screenshot_comment.
                     ScreenshotComment):
                The comment to return text for.

        Returns:
            str:
            The text used to link to the comment.
        """
        assert isinstance(self.obj, Screenshot)
        return self.obj.display_name

    def get_js_model_data(self) -> JSONDict:
        """Return data for the JavaScript model.

        Returns:
            dict:
            Data to serialize for the JavaScript model.
        """
        assert isinstance(self.obj, Screenshot)
        return {
            'imageURL': self.obj.image.url,
            'screenshotID': self.obj.pk,
        }

    def serialize_comments(
        self,
        comments: Sequence[ScreenshotComment],
    ) -> SerializedCommentBlocks[SerializedRegionComment]:
        """Serialize the comments for the screenshot.

        Args:
            comments (list of
                      reviewboard.reviews.models.ScreenshotComment):
                The comments to serialize.

        Returns:
            SerializedCommentBlocks:
            The serialized comments.
        """
        result: SerializedCommentBlocks[SerializedRegionComment] = {}

        for comment in self.flat_serialized_comments(comments):
            position = '%(x)sx%(y)s+%(width)s+%(height)s' % comment
            result.setdefault(position, []).append(comment)

        return result

    def serialize_comment(
        self,
        comment: ScreenshotComment,
    ) -> SerializedRegionComment:
        """Serialize a comment.

        Args:
            comment (reviewboard.reviews.models.ScreenshotComment):
                The comment to serialize.

        Returns:
            SerializedRegionComment:
            The serialized comment.
        """
        return {
            **super().serialize_comment(comment),
            'x': comment.x,
            'y': comment.y,
            'width': comment.w,
            'height': comment.h,
        }



"""Administration form for diff viewer settings."""

from __future__ import annotations

import re
from typing import Any, Dict, List, cast

from django import forms
from django.utils.translation import gettext_lazy as _
from djblets.forms.fields import ListEditDictionaryField
from djblets.forms.widgets import ListEditWidget
from djblets.siteconfig.forms import SiteSettingsForm

from reviewboard.admin.form_widgets import LexersMappingWidget
from reviewboard.codesafety.checkers.trojan_source import \
    TrojanSourceCodeSafetyChecker


class DiffSettingsForm(SiteSettingsForm):
    """Diff settings for Review Board."""

    css_bundle_names = ['djblets-forms']
    js_bundle_names = ['djblets-forms']

    diffviewer_syntax_highlighting = forms.BooleanField(
        label=_('Show syntax highlighting'),
        help_text=_(
            'Show the content of files with colors and formatting to help '
            'reviewers read and follow the code.'
        ),
        required=False)

    diffviewer_syntax_highlighting_threshold = forms.IntegerField(
        label=_('Max lines for syntax highlighting'),
        help_text=_(
            'Files with lines greater than this number will not have '
            'syntax highlighting. Enter 0 to disable limits.'
        ),
        required=False,
        widget=forms.TextInput(attrs={'size': '5'}))

    diffviewer_custom_pygments_lexers = ListEditDictionaryField(
        label=_('Custom file highlighting'),
        help_text=_(
            'Set this to override how particular file extensions (in the '
            'form of <code>.ext</code>) are styled.'
        ),
        required=False,
        widget=ListEditWidget(value_widget=LexersMappingWidget))

    diffviewer_default_tab_size = forms.IntegerField(
        label=_('Tabstop size'),
        help_text=_(
            'Set this to override the default character width for tabstops. '
            'If unset or set to 0, tabstops will show as 8 characters wide.'
        ),
        required=False,
        widget=forms.TextInput(attrs={'size': '2'}))

    diffviewer_show_trailing_whitespace = forms.BooleanField(
        label=_('Show trailing whitespace'),
        help_text=_('Show excess trailing whitespace as red blocks. This '
                    'helps to visualize when a text editor added unwanted '
                    'whitespace to the end of a line.'),
        required=False)

    include_space_patterns = forms.CharField(
        label=_('Show all whitespace for'),
        required=False,
        help_text=_(
            'A comma-separated list of file patterns for which all '
            'whitespace changes should be shown (e.g., "*.py, *.txt"). '
            'This is <strong>no longer recommended</strong>, as it turns off '
            'smart indentation highlighting and can make some changes '
            'harder to review.'
        ),
        widget=forms.TextInput(attrs={'size': '50'}))

    diffviewer_context_num_lines = forms.IntegerField(
        label=_('Lines of context'),
        help_text=_('The number of unchanged lines shown above and below '
                    'changed lines.'),
        initial=5,
        widget=forms.TextInput(attrs={'size': '5'}))

    diffviewer_paginate_by = forms.IntegerField(
        label=_('Paginate by'),
        help_text=_('The number of files to display per page in the diff '
                    'viewer.'),
        initial=20,
        widget=forms.TextInput(attrs={'size': '5'}))

    diffviewer_paginate_orphans = forms.IntegerField(
        label=_('Paginate orphans'),
        help_text=_('The number of extra files required before adding another '
                    'page to the diff viewer.'),
        initial=10,
        widget=forms.TextInput(attrs={'size': '5'}))

    diffviewer_max_diff_size = forms.IntegerField(
        label=_('Max diff size in bytes'),
        help_text=_(
            'The maximum size (in bytes) for any given diff. Enter 0 to '
            'disable size restrictions. <strong>2097152 (2MB) '
            'is recommended</strong>, as larger diffs usually cannot be '
            'reviewed by humans and may cause performance problems and '
            'browser timeouts.'
        ),
        widget=forms.TextInput(attrs={'size': '15'}))

    diffviewer_max_binary_size = forms.IntegerField(
        label=_('Max size for binary files in diffs'),
        help_text=_(
            'The maximum size (in bytes) for any binary files which will be '
            'uploaded to show as part of diffs. <strong>10485760 (10MB) '
            'is recommended</strong>.'),
        widget=forms.TextInput(attrs={'size': '15'}))

    trojan_source_check_confusables = forms.BooleanField(
        label=_(
            'Check for potentially misleading Unicode characters '
            '("confusables")'
        ),
        help_text=_(
            'This will check for Unicode characters in various languages '
            'that look similar to characters commonly used in source code '
            '(such as Latin characters). These are also known as Unicode '
            '"confusables", and can be used to sneak malicious code into '
            'software, either intentionally or accidentally through '
            'copy/paste.'
        ),
        initial=True,
        required=False)

    trojan_source_confusable_aliases_allowed = forms.MultipleChoiceField(
        label=_('Safe character sets'),
        help_text=_(
            'Characters from these sets are considered safe, and will not be '
            'checked for confusables. This is useful if your team is working '
            'in certain languages that often show up as unsafe. '
            '<em>Attacks using these characters will not be caught.</em>'
        ),
        choices=[
            (_lang, _lang)
            for _lang in (
                TrojanSourceCodeSafetyChecker.get_main_confusable_aliases()
            )
        ],
        required=False,
        widget=forms.widgets.CheckboxSelectMultiple())

    def load(self) -> None:
        """Load settings from the form.

        This will populate initial fields based on the site configuration.
        """
        super().load()

        # Load the settings from the Trojan Code checker.
        #
        # In the future, we may want to expand this to dynamically support
        # any and all registered code safety checkers, but that will require
        # additional support in the checkers.
        code_safety_config = cast(
            Dict[str, Dict[str, Any]],
            self.get_key_value('code_safety_checkers'))
        trojan_source_config = code_safety_config.get(
            TrojanSourceCodeSafetyChecker.checker_id, {})

        if trojan_source_config:
            for key in ('check_confusables',
                        'confusable_aliases_allowed'):
                if key in trojan_source_config:
                    self.fields[f'trojan_source_{key}'].initial = \
                        trojan_source_config[key]

        # Load the "Show all whitespace for" setting.
        self.fields['include_space_patterns'].initial = ', '.join(
            cast(List[str],
                 self.get_key_value('diffviewer_include_space_patterns')))

    def save(self) -> None:
        """Save the form.

        This will write the new configuration to the database.
        """
        # Store the settings for the Trojan Code checker.
        code_safety_config = cast(
            Dict[str, Any],
            self.get_key_value('code_safety_checkers'))
        code_safety_config[TrojanSourceCodeSafetyChecker.checker_id] = {
            key: self.cleaned_data[f'trojan_source_{key}']
            for key in ('check_confusables',
                        'confusable_aliases_allowed')
        }
        self.set_key_value('code_safety_checkers', code_safety_config)

        # Save the "Show all whitespace for" setting.
        self.set_key_value(
            'diffviewer_include_space_patterns',
            re.split(r',\s*', self.cleaned_data['include_space_patterns']))

        super().save()

    class Meta:
        """Metadata for the form."""

        title = _('Diff Viewer Settings')

        save_blacklist = (
            'trojan_source_check_confusables',
            'trojan_source_confusable_aliases_allowed',
            'include_space_patterns',
        )

        fieldsets = (
            {
                'title': _('Appearance'),
                'classes': ('wide',),
                'fields': (
                    'diffviewer_show_trailing_whitespace',
                    'diffviewer_syntax_highlighting',
                    'diffviewer_custom_pygments_lexers',
                    'diffviewer_default_tab_size',
                ),
            },
            {
                'title': _('Limits'),
                'classes': ('wide',),
                'description': _(
                    'Limits can be placed to keep large diffs or large files '
                    'from impacting performance.'
                ),
                'fields': (
                    'diffviewer_max_diff_size',
                    'diffviewer_max_binary_size',
                    'diffviewer_syntax_highlighting_threshold',
                ),
            },
            {
                'title': _('Code Safety'),
                'description': _(
                    'Review Board by default checks code for suspicious '
                    'Unicode characters used in '
                    '<a href="https://trojansource.codes/">Trojan Source</a> '
                    'attacks. These checks can be fine-tuned to avoid '
                    'matching characters in some languages, at the expense '
                    'of decreased code safety.'
                ),
                'classes': ('wide',),
                'fields': (
                    'trojan_source_check_confusables',
                    'trojan_source_confusable_aliases_allowed',
                ),
            },
            {
                'title': _('Advanced'),
                'description': _(
                    'These are advanced settings that control the behavior '
                    'and display of the diff viewer. In general, these '
                    'settings do not need to be changed.'
                ),
                'classes': ('wide',),
                'fields': (
                    'include_space_patterns',
                    'diffviewer_context_num_lines',
                    'diffviewer_paginate_by',
                    'diffviewer_paginate_orphans',
                ),
            },
        )
